# node-tair

[![NPM](https://nodei.co/npm/tair.png?downloads=true&stars=true)](https://nodei.co/npm/tair/)

![logo](https://raw.github.com/TBEDP/node-tair/master/logo.png)

Taobao [Tair](http://code.taobao.org/p/tair/src/) Client for Node.js, build on pure javascript without any native code.

Key format following `Java` client.

* Tair [FAQ](http://baike.corp.taobao.com/index.php/Faq)
* [java mc client](http://baike.corp.taobao.com/index.php/CS_RD/tair/mc_java_client)
* [Monitor](http://monitor.tair.taobao.org:9999/)

## Installation

```bash
$ npm install tair
```

Work on `node >= 0.8.0`.

## Usage

```js
var Tair = require('tair');

var tair = Tair({
  dataId: 'market2-daily',
  diamond: 'commonconfig.taobao.net:8080',
  namespace: 171
});

tair.on('connect', function () {
  tair.put('foo', 'bar', function (err) {
    should.not.exist(err);
  });

  tair.get('foo', function (err, data) {
    data.toString().should.equal('bar');
  });
});
```

## API

```js
/**
 * Tair Cluster Client
 *
 * If you are using diamond, you need to set dataId and diamond host.
 * If you are not use diamond, you must set group and servers.
 *
 * @param {Object} options
 *  - {String} dataId diamond data id
 *  - {String} diamond daily: "commonconfig.taobao.net:8080", online: "commonconfig.config-host.taobao.com:8080"
 *  - {Object} [diamondOptions]
 *   - {Number} [refreshInterval] config sync interval, default is 30000 ms, 30 seconds
 *   - {String} [cacheDir] config data cache dir, default is '/tmp/nodejs-diamond-client-cache'
 *   - {Number} [requestTimeout] diamond request timeout, default is 5000 ms
 *
 *  - {String} group group name of Tair cluster
 *  - {Array} servers server list, like `[{"host":"10.235.144.195","port":5198}, ...]`
 *
 *  - {Number} [namespace] default is 0
 *  - {Number} [timeout] request timeout, default is 5000 ms
 *  - {Number} [heartbeatInterval] default is 30000 ms
 *  - {Number} [syncInterval] sync group info interval, default is 30000 ms
 */
function Client(options);

/**
 * Put a value.
 *
 * @param {String} key
 * @param {Buffer|String} value
 * @param {Object} [options]
 *  - {Number} [expired] expired time, default is 0, never expired.
 *  - {Number} [version] default is 0
 * @param {Function(err, res)} callback
 * @return {Client} this
 */
Client.prototype.put = function (key, value, options, callback);

/**
 * Get a value.
 *
 * @param {String} key
 * @param {Object} [options]
 * @param {Function(err, value)} callback
 * @return {Client} this
 */
Client.prototype.get = function (key, options, callback);

/**
 * Prefix get a value.
 *
 * @param {String} pkey primary key
 * @param {String} skey second key
 * @param {Object} [options]
 *  - {Number} [version] default is 0
 * @param {Function(err, value)} callback
 * @return {Client} this
 */
Client.prototype.prefixGet = function (pkey, skey, options, callback);

/**
 * Prefix get mutil values.
 *
 * @param {String} pkey primary key
 * @param {Array} skeys second key list
 * @param {Object} [options]
 *  - {Number} [version] default is 0
 * @param {Function(err, obj)} callback
 *  - {Object} obj
 * @return {Client} this
 */
Client.prototype.prefixGets = function (pkey, skeys, options, callback);

/**
 * Prefix put mutil key values.
 *
 * @param {String|Number|Buffer} pkey primary key
 * @param {Array} keyValues e.g.: `[ ['k1', 'v2'], ['k2', 222] ]`
 *  - {Array} kv `[key, value, version, expire]`, optional: version, expire
 * @param {Object} [options]
 *  - {Number} [namespace] default is `client namespace`
 * @param {Function(err, success)} callback
 *  - {Boolean} success or not
 * @return {Client} this
 */
Client.prototype.prefixPuts = function (pkey, keyValues, options, callback);

/**
 * Invalid a key
 *
 * @param {String|Buffer} key
 * @param {Object} [options]
 * @param {Function(err, success)} callback
 * @return {Client} this
 */
Client.prototype.invalid = function (key, options, callback);

/**
 * 范围查询接口，支持反向查询，返回key, value。
 *
 * @param {String|Number|Buffer} pkey primary key
 * @param {String|Number|Buffer} start start key `start <= value < end`
 * @param {String|Number|Buffer} end end key
 * @param {Number} offset default is 0
 * @param {Number} limit default is 1000
 * @param {Object} [options]
 *  - {Number} namespace
 * @param {Function(err, values)} callback
 * @return {Client} this
 */
Client.prototype.getRange = function (pkey, start, end, offset, limit, options, callback);
Client.prototype.getRangeOnlyValue = function (pkey, start, end, offset, limit, options, callback);
Client.prototype.getRangeOnlyKey = function (pkey, start, end, offset, limit, options, callback);

/**
 * 增加计数。注意：incr前不要put！！
 *
 * @param {String|Number|Buffer} key 不超过1k
 * @param {Number} value 增加量
 * @param {Object} [options]
 *  - {Number} namespace 申请时分配的namespace
 *  - {Number} defaultValue 第一次调用incr时的key的count初始值，第一次返回的值为defaultValue + value
 *  - {Number} expired 数据过期时间，单位为秒，可设相对时间或绝对时间(Unix时间戳)。
 *      expireTime = 0，表示数据永不超时。
 *      expireTime > 0，表示设置超时时间。
 *      若expireTime>当前时间的时间戳，则表示使用绝对时间，否则使用相对时间。
 *      expireTime < 0，表示不关注超时时间，若之前设过超时时间，则已之前的超时时间为准，
 *        若没有，则作为永不超时处理。但当前mdb统一当做永不超时来处理。
 * @param {Function(err, result)} callback
 * @return {Client} this
 */
Client.prototype.incr = function (key, value, options, callback);

/**
 * 增加计数。注意：incr前不要put！！
 *
 * @param {String|Number|Buffer} key 不超过1k
 * @param {Number} value 增加量
 * @param {Object} [options]
 *  - {Number} namespace 申请时分配的namespace
 *  - {Number} defaultValue 第一次调用incr时的key的count初始值，第一次返回的值为defaultValue + value
 *  - {Number} expired 数据过期时间，单位为秒，可设相对时间或绝对时间(Unix时间戳)。
 *      expireTime = 0，表示数据永不超时。
 *      expireTime > 0，表示设置超时时间。
 *      若expireTime>当前时间的时间戳，则表示使用绝对时间，否则使用相对时间。
 *      expireTime < 0，表示不关注超时时间，若之前设过超时时间，则已之前的超时时间为准，
 *        若没有，则作为永不超时处理。但当前mdb统一当做永不超时来处理。
 * @param {Function(err, result)} callback
 * @return {Client} this
 */
Client.prototype.incr = function (key, value, options, callback);

/**
 * 设置计数。设置某一个key的当前计数，可以用于incr/decr之前或者之后。
 *
 * @see http://baike.corp.taobao.com/index.php/Faq#.E4.BB.80.E4.B9.88.E6.98.AFversion.EF.BC.9F
 * @param {String|Number|Buffer} key
 * @param {Number} count
 * @param {Object} [options]
 *  - {Number} namespace 申请时分配的namespace
 *  - {Number} version 为了解决并发更新同一个数据而设置的参数。默认为 0
 *      当version为0时，表示强制更新。当传入get value返回的版本号，
 *      则若此key未发生更新时，返回正常，否则返回version error。version具体原理和使用方法请参见什么是version。
 *  - {Number} expired 数据过期时间，单位为秒，可设相对时间或绝对时间(Unix时间戳)。
 *      expireTime = 0，表示数据永不超时。
 *      expireTime > 0，表示设置超时时间。
 *      若expireTime>当前时间的时间戳，则表示使用绝对时间，否则使用相对时间。
 *      expireTime < 0，表示不关注超时时间，若之前设过超时时间，则已之前的超时时间为准，
 *        若没有，则作为永不超时处理。但当前mdb统一当做永不超时来处理。
 * @param {Function(err, result)} callback
 * @return {Client} this
 */
Client.prototype.setCount = function (key, count, options, callback);

/**
 * Lock a key.
 *
 * @param {String|Number|Buffer} key
 * @param {Number} expired lock expried in seconds.
 * @param {Object} [options]
 *  - {Number} [namespace] default is global namespace.
 * @param {Function(err, success, res)} callback
 * @return {Client}
 */
Client.prototype.lock = function (key, expired, options, callback);

/**
 * unlock a key
 *
 * @param {String|Number|Buffer} key
 * @param {Object} [options]
 *  - {Number} [namespace] default is global namespace.
 * @param {Function(err, success, res)} callback
 * @return {Client} this
 */
Client.prototype.unlock = function (key, options, callback);
```

## TODO

* [ ] prefix APIs
    * [x] prefixGet(pkey, skey)
    * [x] prefixPut(pkey, skey)
    * [ ] prefixInvalid(pkey, skey)
    * [x] prefixGets(pkey, skeys)
    * [x] prefixPuts(pkey, keyvalues)
    * [ ] prefixInvailds(pkey, skeys)
    * [ ] prefixIncr(pkey, skey, value)
    * [ ] prefixDecr(pkey, skey, value)
    * [ ] prefixSetCount(pkey, skey, count)
    * [ ] prefixGetHidden(pkey, skey)
    * [ ] prefixGetHides(pkey, skeys)
    * [ ] prefixHideByProxy(pkey, skey)
    * [ ] prefixHidesByProxy(pkey, skeys)
* [x] put
* [x] get
* [x] invaild
* [ ] mget
* [ ] mput
* [x] lock(key, expired, options)
* [x] unlock(key, options)
* [x] getRange(pkey, start, end) (TODO: reverse not support)
    * [x] getRange(pkey, start, end)
    * [x] getRangeOnlyValue(pkey, start, end)
    * [x] getRangeOnlyKey(pkey, start, end)
* [x] incr(key, value)
* [x] decr(key, value)
* [x] setCount(key, count)
* [ ] hideByProxy(key)
* [ ] getHidden(key)

## 两个集群单备份的容灾流程

* 从 diamond 读取 'dataId' + '-GROUP' 的集群列表信息, 一般会有两个集群 (双机房), 如

```
10.1.=ldbcommon-daily-1
10.2.=ldbcommon-daily-2
```

* 根据当前机器 ip 前缀, 选择跟ip最匹配的一个集群, 并且将另外一个集群作为备份集群, 当默认集群无节点的时候, 从备份集群读取数据(跨机房).

* 继续去 diamond 读取集群的 master 和 slave 配置信息 `_handleClusterConfig()`

* 去 master 或者 slave 读取当前集群的节点信息 `_syncGroupInfo()`

* 定时同步备份集群的节点信息, 用于默认集群出现异常后的容灾方案. (目前只读, 不写)

## Benchmark result

More detail [see here](https://github.com/TBEDP/node-tair/tree/master/benchmark).

### Scripts

```bash
$ node benchmark/bench.js [get|set] [bytes=100] [concurrency=10] [requests=10000]
```

### Result table

Method | Concurrency | Total | Bytes | QPS      | rt avg (ms) | rt max | rt min
---    | ---         | ---   | ---   | ---      | ---    | ---    | ---
get    | 1           | 10000 | 100   | 716.076 | 1.393 | 13.466 | 0.611
get    | 2           | 20000 | 100   | 1486.436 | 1.338 | 12.417 | 0.597
get    | 3           | 30000 | 100   | 2682.164 | 1.114 | 13.67 | 0.593
get    | 4           | 40000 | 100   | 3054.135 | 1.304 | 13.659 | 0.594
get    | 5           | 50000 | 100   | 3143.863 | 1.583 | 13.5 | 0.593
get    | 6           | 60000 | 100   | 3774.534 | 1.583 | 16.788 | 0.582
get    | 7           | 70000 | 100   | 3682.854 | 1.892 | 15.803 | 0.637
get    | 8           | 80000 | 100   | 3934.104 | 2.028 | 16.045 | 0.638
get    | 9           | 90000 | 100   | 4151.100 | 2.162 | 25.832 | 0.635
get    | 10          | 100000 | 100   | 4164.064 | 2.392 | 22.737 | 0.611
get    | 15          | 150000 | 100   | 4567.600 | 3.274 | 23.662 | 0.676
get    | 20          | 200000 | 100   | 4592.317 | 4.340 | 34.337 | 0.684
get    | 30          | 300000 | 100   | 4606.597 | 6.492 | 30.432 | 0.654
get    | 50          | 500000 | 100   | 4680.728 | 10.625 | 201.956 | 0.693
put    | 1           | 10000 | 100   | 1023.332 | 0.974 | 13.812 | 0.604
put    | 2           | 20000 | 100   | 1429.286 | 1.394 | 12.911 | 0.579
put    | 3           | 30000 | 100   | 2629.042 | 1.136 | 203.712 | 0.614
put    | 4           | 40000 | 100   | 2582.478 | 1.545 | 13.612 | 0.593
put    | 5           | 50000 | 100   | 2937.720 | 1.691 | 17.701 | 0.609
put    | 6           | 60000 | 100   | 3580.807 | 1.669 | 209.839 | 0.652
put    | 7           | 70000 | 100   | 3578.549 | 1.946 | 15.566 | 0.651
put    | 8           | 80000 | 100   | 3898.445 | 2.045 | 29.572 | 0.664
put    | 9           | 90000 | 100   | 4195.217 | 2.137 | 18.582 | 0.669
put    | 10          | 100000 | 100   | 4018.324 | 2.477 | 20.385 | 0.672
put    | 15          | 150000 | 100   | 4444.840 | 3.361 | 19.013 | 0.683
put    | 20          | 200000 | 100   | 4351.515 | 4.578 | 33.545 | 0.65
put    | 30          | 300000 | 100   | 4689.112 | 6.374 | 29.206 | 0.684
put    | 50          | 500000 | 100   | 4638.047 | 10.735 | 44.306 | 0.697

## Tair Server API

实现 Tair 通信协议的 Server.

```js
var tair = require('tair');

var configServer1 = tair.createConfigServer();
configServer1.listen(5198);

var configServer2 = tair.createConfigServer();
configServer2.listen(5199);

var configServers = [
  {host: '127.0.0.1', port: 5198},
  {host: '127.0.0.1', port: 5199}
];
var dataNode1 = tair.createDataNode({
  configServers: configServers,
  group: 'testgroup1'
});
var dataNode2 = tair.createDataNode({
  configServers: configServers,
  group: 'testgroup1'
});
var dataNode3 = tair.createDataNode({
  configServers: configServers,
  group: 'testgroup1'
});

function routes(datanode) {
  datanode.get(500, function (req, res, next) {
    if (req.data.key === '404') {
      return next();
    }
    if (req.data.key === '500') {
      return next({message: 'mock error', status: 500});
    }
    cache.get(req.data.key, function (err, value) {
      if (err) {
        return next(err);
      }

      res.send(value);
    });
  });
}

routes(dataNode1);
routes(dataNode2);
routes(dataNode3);

dataNode1.listen(5192);
dataNode2.listen(5193);
dataNode3.listen(5194);
```

## Authors

```bash
$ git summary

 project  : node-tair
 repo age : 1 year, 5 months
 active   : 70 days
 commits  : 204
 files    : 78
 authors  :
   115  fengmk2                 56.4%
    78  sunfang1cn              38.2%
     6  sun fang                2.9%
     2  卡特                  1.0%
     1  dead-horse              0.5%
     1  fish                    0.5%
     1  老A                    0.5%
```

## License

(The MIT License)

Copyright (c) 2012 - 2013 Alibaba Group Holding Limited

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the 'Software'), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
