/**!
 * node-tair - lib/client.js
 *
 * Copyright(c) 2012 - 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 *   sunfang1cn <kate.sf@taobao.com>
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:client');
var EventEmitter = require('events').EventEmitter;
var util = require('util');
var packets = require('./packets');
var Connection = require('./connection');
var murmurhash2 = require('./utils').murmurhash2;
var Transcoder = require('./packets/transcoder');
var cons = require('./const');
var MixedKey = packets.MixedKey;

/**
 * Tair client
 *
 * If you are using diamond, please use Cluster mode.
 *
 * @param {Object} options
 *  - {String} group group name of Tair cluster
 *  - {Array} servers server list, like `[{"host":"10.235.144.195","port":5198}, ...]`
 *  - {Number} [namespace] default is 0
 *  - {Number} [timeout] request timeout, default is 5000 ms
 *  - {Number} [heartbeatInterval] default is 30000 ms
 *  - {Number} [syncInterval] sync group info interval, default is 30000 ms
 *  - {Boolean} [withHeader] true: java format, false: cpp format. default is true.
 */
function Client(options) {
  EventEmitter.call(this);
  this._namespace = options.namespace || 0;
  this._timeout = options.timeout || 5000;
  this._heartbeatInterval = options.heartbeatInterval || 30000;
  this._syncInterval = options.syncInterval || 30000;
  this._version = 0;
  this._debug = options.debug;
  this._timer = options.timer || Date;
  this.logger = options.logger || console;

  this._clusterName = options.clusterName || 'default-cluster';
  this._groupName = options.group;
  // group info
  this._groupInfo = {
    bucket: 0,
    copy: 0,
    nodes: {},
    nodeCount: 0,
    servers: [],
    version: 0,
    configCount: 0,
    config: {},
    invalidateServers: [],
    downServerCount: 0,
    downServers: {}
  };
  this._servers = options.servers || [];
  this._invalidateServerIndex = 0;

  this._connections = {};
  this._transcoder = new Transcoder(options.withHeader);

  this._syncTimer = null;
  this._init();
}

util.inherits(Client, EventEmitter);

/**
 * APIs
 */

/**
 * Put a value.
 *
 * @param {String} key
 * @param {Buffer|String} value
 * @param {Object} [options]
 *  - {Number} [expired] expired time, default is 0, never expired.
 *  - {Number} [version] default is 0
 *  - {Number} [namespace]
 * @param {Function(err, res)} callback
 * @return {Client} this
 */
Client.prototype.put = function (key, value, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }
  options = options || {};
  options.namespace = options.namespace || this._namespace;
  var put = new packets.Put(key, value, this._transcoder, options);
  this._send(put, function (err, body, res) {
    if (err) {
      return callback(err);
    }
    callback(null, body && body.code >= 0, res);
  });
  return this;
};
Client.prototype.set = Client.prototype.put;

/**
 * Get a value.
 *
 * @param {String} key
 * @param {Object} [options]
 *  - {Number} [namespace]
 * @param {Function(err, value)} callback
 * @return {Client} this
 */
Client.prototype.get = function (key, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }

  options = options || {};
  options.namespace = options.namespace || this._namespace;
  var get = new packets.Get(key, this._transcoder, options);
  this._send(get, function (err, body, res) {
    if (err) {
      return callback(err, body, res);
    }
    callback(null, body.value, res);
  });
  return this;
};

Client.prototype._delete = function (key, options, callback) {
  var del = new packets.Remove(key, this._transcoder, options);
  this._send(del, function (err, body, res) {
    if (err) {
      return callback(err, body, res);
    }
    callback(null, body.code >= 0, res);
  });
  return this;
};

/**
 * Invalid a key
 *
 * @param {String|Buffer} key
 * @param {Object} [options]
 *  - {Number} [namespace]
 * @param {Function(err, success)} callback
 * @return {Client} this
 */
Client.prototype.invalid = function (key, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }

  options = options || {};
  options.namespace = options.namespace || this._namespace;
  var node = this._getInvalidateNode();
  if (!node) {
    return this._delete(key, options, callback);
  }

  var invalid = new packets.Invalid(key, this._groupName, this._transcoder, options);
  this._send(invalid, node, function (err, body, res) {
    if (err) {
      return callback(err);
    }
    callback(null, body && body.code >= 0, res);
  });
  return this;
};

Client.prototype._addCount = function (key, value, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }
  options = options || {};
  options.namespace = options.namespace || this._namespace;
  options.defaultValue = options.defaultValue || 0;
  options.version = options.version || 0;
  options.expired = options.expired || 0;

  var incdec = new packets.IncDec(key, value, this._transcoder, options);
  this._send(incdec, function (err, body, res) {
    if (err) {
      return callback(err, body, res);
    }
    callback(null, body.value, res);
  });
  return this;
};

/**
 * 设置计数。设置某一个key的当前计数，可以用于incr/decr之前或者之后。
 *
 * @see http://baike.corp.taobao.com/index.php/Faq#.E4.BB.80.E4.B9.88.E6.98.AFversion.EF.BC.9F
 * @param {String|Number|Buffer} key
 * @param {Number} count
 * @param {Object} [options]
 *  - {Number} namespace 申请时分配的namespace
 *  - {Number} version 为了解决并发更新同一个数据而设置的参数。默认为 0
 *      当version为0时，表示强制更新。当传入get value返回的版本号，
 *      则若此key未发生更新时，返回正常，否则返回version error。version具体原理和使用方法请参见什么是version。
 *  - {Number} expired 数据过期时间，单位为秒，可设相对时间或绝对时间(Unix时间戳)。
 *      expireTime = 0，表示数据永不超时。
 *      expireTime > 0，表示设置超时时间。
 *      若expireTime>当前时间的时间戳，则表示使用绝对时间，否则使用相对时间。
 *      expireTime < 0，表示不关注超时时间，若之前设过超时时间，则已之前的超时时间为准，
 *        若没有，则作为永不超时处理。但当前mdb统一当做永不超时来处理。
 * @param {Function(err, result)} callback
 * @return {Client} this
 */
Client.prototype.setCount = function (key, count, options, callback) {
  var value = new packets.IncData(count);
  return this.put(key, value, options, callback);
};

/**
 * 增加计数。注意：incr前不要put！！
 *
 * @param {String|Number|Buffer} key 不超过1k
 * @param {Number} value 增加量
 * @param {Object} [options]
 *  - {Number} namespace 申请时分配的namespace
 *  - {Number} defaultValue 第一次调用incr时的key的count初始值，第一次返回的值为defaultValue + value
 *  - {Number} expired 数据过期时间，单位为秒，可设相对时间或绝对时间(Unix时间戳)。
 *      expireTime = 0，表示数据永不超时。
 *      expireTime > 0，表示设置超时时间。
 *      若expireTime>当前时间的时间戳，则表示使用绝对时间，否则使用相对时间。
 *      expireTime < 0，表示不关注超时时间，若之前设过超时时间，则已之前的超时时间为准，
 *        若没有，则作为永不超时处理。但当前mdb统一当做永不超时来处理。
 * @param {Function(err, result)} callback
 * @return {Client} this
 */
Client.prototype.incr = function (key, value, options, callback) {
  return this._addCount(key, value, options, callback);
};

/**
 * 减少计数。注意：decr前不要put！！
 *
 * @param {String|Number|Buffer} key
 * @param {Number} value 增加量
 * @param {Object} [options]
 *  - {Number} namespace 申请时分配的namespace
 *  - {Number} defaultValue 第一次调用incr时的key的count初始值，第一次返回的值为defaultValue + value
 *  - {Number} expired 数据过期时间，单位为秒，可设相对时间或绝对时间(Unix时间戳)。
 *      expireTime = 0，表示数据永不超时。
 *      expireTime > 0，表示设置超时时间。
 *      若expireTime>当前时间的时间戳，则表示使用绝对时间，否则使用相对时间。
 *      expireTime < 0，表示不关注超时时间，若之前设过超时时间，则已之前的超时时间为准，
 *        若没有，则作为永不超时处理。但当前mdb统一当做永不超时来处理。
 * @param {Function(err, result)} callback
 * @return {Client} this
 */
Client.prototype.decr = function (key, value, options, callback) {
  return this._addCount(key, -value, options, callback);
};

Client.prototype.close = function () {
  for (var k in this._connections) {
    var conn = this._connections[k];
    conn.close();
  }
  this._connections = {};
  setTimeout(function () {
    debug('Tair %s:%s client closed.', this._clusterName, this._groupName);
    this.emit('close');
  }.bind(this), 500);
  if (this._syncTimer) {
    clearInterval(this._syncTimer);
    this._syncTimer = null;
  }
  return this;
};

/**
 * Prefix APIs
 */

/**
 * Prefix put a value.
 *
 * @param {String} pkey primary key
 * @param {String} skey second key
 * @param {Buffer|String} value
 * @param {Object} [options]
 *  - {Number} [expired] expired time, default is 0, never expired.
 *  - {Number} [version] default is 0
 * @param {Function(err, res)} callback
 * @return {Client} this
 */
Client.prototype.prefixPut = function (pkey, skey, value, options, callback) {
  var key = new MixedKey(this._transcoder, pkey, skey);
  return this.put(key, value, options, callback);
};

/**
 * Prefix get a value.
 *
 * @param {String} pkey primary key
 * @param {String} skey second key
 * @param {Object} [options]
 *  - {Number} [version] default is 0
 * @param {Function(err, value)} callback
 * @return {Client} this
 */
Client.prototype.prefixGet = function (pkey, skey, options, callback) {
  var key = new MixedKey(this._transcoder, pkey, skey);
  return this.get(key, options, callback);
};

/**
 * Prefix get mutil values.
 *
 * @param {String} pkey primary key
 * @param {Array} skeys second key list
 * @param {Object} [options]
 *  - {Number} [version] default is 0
 * @param {Function(err, obj)} callback
 *  - {Object} obj
 * @return {Client} this
 */
Client.prototype.prefixGets = function (pkey, skeys, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }

  options = options || {};
  options.namespace = options.namespace || this._namespace;

  var keys = [];
  for (var i = 0; i < skeys.length; i++) {
    var key = new MixedKey(this._transcoder, pkey, skeys[i]);
    keys.push(key);
  }
  var get = new packets.PrefixGet(keys, this._transcoder, options);
  this._send(get, function (err, body, res) {
    if (err) {
      return callback(err, body, res);
    }
    callback(null, body.value, res);
  });
  return this;
};

/**
 * Prefix put mutil key values.
 *
 * @param {String|Number|Buffer} pkey primary key
 * @param {Array} keyValues e.g.: `[ ['k1', 'v2'], ['k2', 222] ]`
 *  - {Array} kv `[key, value, version, expire]`, optional: version, expire
 * @param {Object} [options]
 *  - {Number} [namespace] default is `client namespace`
 * @param {Function(err, success)} callback
 *  - {Boolean} success or not
 * @return {Client} this
 */
Client.prototype.prefixPuts = function (pkey, keyValues, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }

  options = options || {};
  options.namespace = options.namespace || this._namespace;

  var puts = new packets.PrefixPuts(pkey, keyValues, this._transcoder, options);
  this._send(puts, function (err, body, res) {
    if (err) {
      return callback(err, body, res);
    }
    callback(null, body.code >= 0, res);
  });
  return this;
};

/**
 * Lock a key.
 *
 * @param {String|Number|Buffer} key
 * @param {Number} expired lock expried in seconds.
 * @param {Object} [options]
 *  - {Number} [namespace] default is global namespace.
 * @param {Function(err, success, res)} callback
 * @return {Client}
 */
Client.prototype.lock = function (key, expired, options, callback) {
  // 如果key 不存在, 则先put 然后再 lock
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }
  expired = Number(expired);
  if (!expired || expired <= 0) {
    expired = 60;
  }
  options = options || {};
  options.namespace = options.namespace || this._namespace;
  var lock = new packets.Lock(key, packets.Lock.LOCK_VALUE, this._transcoder, options);
  var that = this;
  that._doLock(lock, function (err, success, res) {
    if (success) {
      return callback(null, success, res);
    }

    if (err && (err.code === -3998 || err.code === -3988)) {
      // TairRequestError: -3998: data not exist
      // create the key and lock again
      var putOptions = {
        expired: expired,
        namespace: options.namespace,
      };
      that.put(key, key + ' lock', putOptions, function (err, result, res) {
        if (err) {
          return callback(err, result, res);
        }
        var putUse = res && res.use || 0;
        // lock again
        that._doLock(lock, function (err, success, res) {
          res = res || {};
          res.use = res.use + putUse;
          callback(err, success, res);
        });
      });
      return;
    }

    callback(err, success, res);
  });
  return this;
};

/**
 * unlock a key
 *
 * @param {String|Number|Buffer} key
 * @param {Object} [options]
 *  - {Number} [namespace] default is global namespace.
 * @param {Function(err, success, res)} callback
 * @return {Client} this
 */
Client.prototype.unlock = function (key, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }
  options = options || {};
  options.namespace = options.namespace || this._namespace;
  var lock = new packets.Lock(key, packets.Lock.UNLOCK_VALUE, this._transcoder, options);
  this._doLock(lock, function (err, success, res) {
    if (err && (err.code === -3974 || err.code === -3988)) {
      // ignore lock not exists
      err = null;
      success = true;
    }
    callback(err, success, res);
  });
  return this;
};

Client.prototype._doLock = function (lock, callback) {
  this._send(lock, function (err, body, res) {
    if (err) {
      return callback(err, false, res);
    }
    callback(null, body.code >= 0, res);
  });
};

/**
 * 范围查询接口，支持反向查询，返回key, value。
 *
 * @param {String|Number|Buffer} pkey primary key
 * @param {String|Number|Buffer} start start key `start <= value < end`
 * @param {String|Number|Buffer} end end key
 * @param {Number} offset default is 0
 * @param {Number} limit default is 1000
 * @param {Object} [options]
 *  - {Number} namespace
 * @param {Function(err, values)} callback
 * @return {Client} this
 */
Client.prototype.getRange = function (pkey, start, end, offset, limit, options, callback) {
  return this._getRangeCmd(pkey, start, end, offset, limit, cons.CMD_RANGE_ALL, options, callback);
};

/**
 * 范围查询接口，支持反向查询，返回key
 *
 * @param {String|Number|Buffer} pkey primary key
 * @param {String|Number|Buffer} start start key `start <= value < end`
 * @param {String|Number|Buffer} end end key
 * @param {Number} offset default is 0
 * @param {Number} limit default is 1000
 * @param {Object} [options]
 *  - {Number} namespace
 * @param {Function(err, values)} callback
 * @return {Client} this
 */
Client.prototype.getRangeOnlyKey = function (pkey, start, end, offset, limit, options, callback) {
  return this._getRangeCmd(pkey, start, end, offset, limit, cons.CMD_RANGE_KEY_ONLY, options, callback);
};

/**
 * 范围查询接口，支持反向查询，返回value。
 *
 * @param {String|Number|Buffer} pkey primary key
 * @param {String|Number|Buffer} start start key `start <= value < end`
 * @param {String|Number|Buffer} end end key
 * @param {Number} offset default is 0
 * @param {Number} limit default is 1000
 * @param {Object} [options]
 *  - {Number} namespace
 * @param {Function(err, values)} callback
 * @return {Client} this
 */
Client.prototype.getRangeOnlyValue = function (pkey, start, end, offset, limit, options, callback) {
  return this._getRangeCmd(pkey, start, end, offset, limit, cons.CMD_RANGE_VALUE_ONLY, options, callback);
};

/**
 * Internal methods
 */

Client.prototype._getRangeCmd = function (pkey, start, end, offset, limit, cmdType, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }
  options = options || {};
  options.namespace = options.namespace || this._namespace;
  if (!limit) {
    limit = cons.DEFAULT_RANGE_LIMIT;
  }

  // for Compatibility
  if (start === '') {
    start = null;
  }
  if (end === '') {
    end = null;
  }

  var startKey = new MixedKey(this._transcoder, pkey, start);
  var endKey = new MixedKey(this._transcoder, pkey, end);
  var getRange = new packets.GetRange(startKey, endKey, cmdType, offset, limit, this._transcoder, options);

  this._send(getRange, function (err, body, res) {
    if (err) {
      return callback(err, body, res);
    }
    callback(null, body.items, res);
  });
  return this;
};

Client.prototype.updateGroup = function (group, servers) {
  this._groupName = group;
  this._servers = servers;
};

Client.prototype._init = function () {
  // the first sync success, meaning client connected.
  var that = this;
  that.once('syncSuccess', that.emit.bind(that, 'connect'));

  if (that._servers.length > 0) {
    // no use diamond
    that._syncTimer = setInterval(that._syncGroupInfo.bind(that), that._syncInterval);
    that._syncGroupInfo();
    return;
  }
};

Client.prototype._syncGroupInfo = function () {
  var that = this;
  var groupName = that._groupName;
  var groupInfo = that._groupInfo;
  var clusterName = that._clusterName;
  debug('Syncing %s:%s info...', clusterName, groupName);
  that._getGroup(that._version, groupName, clusterName, function (err, group) {
    if (err) {
      debug('Sync %s:%s info error: %s', clusterName, groupName, err);
      return that.emit('syncError', err);
    }
    for (var k in group) {
      groupInfo[k] = group[k];
    }
    debug('Sync %s:%s info success, %d bucket, %d copy, version %s, config: %j',
      clusterName, groupName, group.bucket, group.copy, group.version, group.config);
    if (!group.invalidateServers || !group.invalidateServers.length) {
      debug('%s:%s no invalidateServers: %j',
        clusterName, groupName, group.invalidateServers);
    }
    that.emit('syncSuccess');
  });
};

Client.prototype._getGroup = function (version, name, clusterName, callback) {
  var getGroup = new packets.GetGroup(version, name, clusterName);
  var that = this;
  var i = 0;
  var lastErr = null;
  var next = function () {
    if (i >= that._servers.length) {
      if (!lastErr) {
        lastErr = new Error('Can not got ' + name + ' group infomation');
        lastErr.name = 'TairGetGroupError';
      }
      return callback(lastErr);
    }
    var server = that._servers[i++];
    var connection = that._getConnection(server.host, server.port);
    connection.call(getGroup, function (err, res) {
      if (err) {
        lastErr = err;
        return next();
      }
      callback(null, res.body);
    });
  };

  next();
};

function connectionHeartbeat() {
  var that = this;
  that.call(new packets.Ping(), function (err, res) {
    if (err || !res || !res.body || res.body.code < 0) {
      console.warn('[%s] [%s] connection:%s ping error, %s, %j',
        Date(), process.pid, that.name, err, res);
      // close the connection.
      that.close();
    }
  });
}

Client.prototype._getConnection = function (host, port) {
  var key = host + ':' + port;
  var conn = this._connections[key];
  if (conn) {
    return conn;
  }

  conn = new Connection({
    host: host,
    port: port,
    timeout: this.timeout,
    debug: this._debug,
    timer: this._timer,
  });
  this._connections[key] = conn;

  conn._heartbeat = connectionHeartbeat;
  conn._heartbeat();
  var heartbeatTimer = setInterval(conn._heartbeat.bind(conn), this._heartbeatInterval);

  // remove it when close event emit.
  conn.once('close', function () {
    delete this._connections[key];
    // remove heartbeat timer
    clearInterval(heartbeatTimer);
    heartbeatTimer = null;
  }.bind(this));
  return conn;
};

Client.prototype._send = function (packet, node, callback) {
  if (typeof node === 'function') {
    callback = node;
    node = null;
  }
  var isMixedKey = false;
  if (!node) {
    isMixedKey = packet.key instanceof MixedKey;
    var keyBytes = isMixedKey ? this._transcoder.encode(packet.key, true) : packet.keyBytes;
    node = this._getDataNode(keyBytes, packet.isRead);
  }
  // debug('get key "%j" node: %j, isMixedKey: %s', packet.key, node, isMixedKey);

  if (node.error) {
    return callback(node.error);
  }

  var connection = this._getConnection(node.host, node.port);
  connection.call(packet, function (err, res) {
    if (err) {
      return callback(err, null, res);
    }
    if (res.body.code < 0) {
      var message = res.body.message || packets.ResultCodes[res.body.code] || 'unknow error';
      err = new Error(res.body.code + ': ' + message);
      err.name = 'TairRequestError';
      err.code = res.body.code;
      err.node = node.host + ':' + node.port;
      err.body = res.body;
      return callback(err, res.body, res);
    }
    callback(null, res.body, res);
  });
};

Client.prototype._getInvalidateNode = function () {
  if (this._invalidateServerIndex >= this._groupInfo.invalidateServers.length) {
    this._invalidateServerIndex = 0;
  }
  return this._groupInfo.invalidateServers[this._invalidateServerIndex++];
};

Client.prototype._getClusterDataNode = function (clusterType, hash, isRead) {
  var clusterName;
  var groupInfo;
  var groupName;
  if (clusterType === 'primary') {
    clusterName = this._clusterName;
    groupInfo = this._groupInfo;
    groupName = this._groupName;
  } else {
    clusterName = this._backupGroup.clusterName;
    groupInfo = this._backupGroup.info;
    groupName = this._backupGroup.name;
    debug('get data node:%s from backup: %s:%s:%s', hash, clusterType, clusterName, groupName);
  }

  var nodeIndex = parseInt(hash %= groupInfo.bucket, 10) || 0;
  var nodes = groupInfo.servers[0] || [];
  var node = nodes[nodeIndex] || {};
  var nodeId = node.host + ':' + node.port;
  var aliveNode = groupInfo.nodes[nodeId];

  if (!aliveNode) {
    this.logger.warn("[%s] [%s] TairDataNodeNotAlive: %s:%s:%s data node:%d(%s:%j) not alive",
      Date(), process.pid, clusterType, clusterName, groupName, nodeIndex, nodeId, node);
  }

  if (aliveNode) {
    var downServerKey = aliveNode.host + ':' + aliveNode.port;
    if (groupInfo.downServers[downServerKey]) {
      aliveNode = null;
      this.logger.warn("[%s] [%s] TairDateNodeDown: %s:%s:%s data node:%d(%s:%j, key: %s) had down",
        Date(), process.pid, clusterType, clusterName, groupName, nodeIndex, nodeId, node, downServerKey);
    }
  }

  // if read operation, try to get data from copy slaves
  if (!aliveNode && isRead && groupInfo.copy > 1) {
    for (var i = 1; i < groupInfo.copy; i++) {
      nodes = groupInfo.servers[i] || [];
      var slave = nodes[nodeIndex] || {};
      var slaveId = slave.host + ':' + slave.port;
      aliveNode = groupInfo.nodes[slaveId];
      if (aliveNode && !groupInfo.downServers[aliveNode.host + ':' + aliveNode.port]) {
        debug("Tair: read operation try on %s:%s:%s copy:%d %s:%j",
          clusterType, clusterName, groupName, i, slaveId, aliveNode);
        break;
      }
      aliveNode = null;
    }
    if (!aliveNode) {
      this.logger.warn("[%s] [%s] TairDataNodeAllDown: %s:%s:%s all slave servers also had down, copy: %s",
        Date(), process.pid, clusterType, clusterName, groupName, groupInfo.copy);
    }
  }

  return aliveNode;
};

Client.prototype._getDataNode = function (keyBytes, isRead) {
  var dataNode = {
    host: '',
    port: 0,
    error: null,
  };

  var hash = murmurhash2(keyBytes);

  var clusterName = this._clusterName;
  var groupInfo = this._groupInfo;
  var groupName = this._groupName;

  var nodeIndex = parseInt(hash %= groupInfo.bucket, 10) || 0;
  var nodes = groupInfo.servers[0] || [];
  var node = nodes[nodeIndex] || {};
  var nodeId = node.host + ':' + node.port;
  var aliveNode = groupInfo.nodes[nodeId];

  if (!aliveNode) {
    this.logger.warn("[%s] [%s] TairDataNodeNotAlive: %s:%s data node:%d(%s:%j) not alive",
      Date(), process.pid, clusterName, groupName, nodeIndex, nodeId, node);
  }

  if (aliveNode) {
    var downServerKey = aliveNode.host + ':' + aliveNode.port;
    if (groupInfo.downServers[downServerKey]) {
      aliveNode = null;
      this.logger.warn("[%s] [%s] TairDateNodeDown: %s:%s data node:%d(%s:%j, key: %s) had down",
        Date(), process.pid, clusterName, groupName, nodeIndex, nodeId, node, downServerKey);
    }
  }

  // if read operation, try to get data from copy slaves
  if (!aliveNode && isRead && groupInfo.copy > 1) {
    for (var i = 1; i < groupInfo.copy; i++) {
      nodes = groupInfo.servers[i] || [];
      var slave = nodes[nodeIndex] || {};
      var slaveId = slave.host + ':' + slave.port;
      aliveNode = groupInfo.nodes[slaveId];
      if (aliveNode && !groupInfo.downServers[aliveNode.host + ':' + aliveNode.port]) {
        debug("Tair: read operation try on %s:%s copy:%d %s:%j",
          clusterName, groupName, i, slaveId, aliveNode);
        break;
      }
      aliveNode = null;
    }
    if (!aliveNode) {
      this.logger.warn("[%s] [%s] TairDataNodeAllDown: %s:%s all slave servers also had down, copy: %s",
        Date(), process.pid, clusterName, groupName, groupInfo.copy);
    }
  }

  if (!aliveNode) {
    dataNode.error = new Error('Tair data node#' + hash + '(' + nodeId + ')' + ' not alive');
    dataNode.error.name = 'TairDataNodeNotAliveError';
  } else {
    dataNode.host = aliveNode.host;
    dataNode.port = aliveNode.port;
  }
  return dataNode;
};

module.exports = Client;
