/**!
 * node-tair - lib/cluster_client.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * Authors:
 *   苏千 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:cluster_client');
var util = require('util');
var utility = require('utility');
var EventEmitter = require('events').EventEmitter;
var address = require('address');
var Diamond = require('diamond-client');
var eventproxy = require('eventproxy');
var Client = require('./client');

/**
 * Tair Cluster Client
 *
 * @param {Object} options
 *  - {String} dataId diamond data id
 *  - {String} diamond daily: "commonconfig.taobao.net:8080", online: "commonconfig.config-host.taobao.com:8080"
 *  - {Object} [diamondOptions]
 *   - {Number} [refreshInterval] config sync interval, default is 30000 ms, 30 seconds
 *   - {String} [cacheDir] config data cache dir, default is '/tmp/nodejs-diamond-client-cache'
 *   - {Number} [requestTimeout] diamond request timeout, default is 5000 ms
 *  - {Number} [namespace] default is 0
 *  - {Number} [timeout] request timeout, default is 5000 ms
 *  - {Number} [heartbeatInterval] default is 30000 ms
 *  - {Number} [syncInterval] sync group info interval, default is 30000 ms
 *  - {Boolean} [withHeader] true: java format, false: cpp format. default is true.
 */

function ClusterClient(options) {
  EventEmitter.call(this);

  this._dataId = options.dataId;
  this._diamondOptions = options.diamondOptions || {};
  this._diamondOptions.host = options.diamond;
  this._diamondOptions.refreshInterval = this._diamondOptions.refreshInterval || 30000;
  this._diamondOptions.requestTimeout = this._diamondOptions.requestTimeout || 5000;

  this._namespace = options.namespace || 0;
  this._timeout = options.timeout || 5000;
  this._heartbeatInterval = options.heartbeatInterval || 30000;
  this._syncInterval = options.syncInterval || 30000;
  this._version = 0;
  this._debug = options.debug;
  this._timer = options.timer || Date;
  this.logger = options.logger || console;
  this._withHeader = options.withHeader;

  this._localIP = address.ip() || '';
  if (!this._localIP) {
    this.logger.warn('[%s] [%s] can not found server local ip', Date(), process.pid);
  }

  // name1, name2, ...
  this._clusterNames = [];
  // name: {client, readWeight, writeWeight, connected}
  this._clusterMap = {};
  this._init();
}

util.inherits(ClusterClient, EventEmitter);
module.exports = ClusterClient;

ClusterClient.prototype._init = function () {
  // the first sync success, meaning client connected.
  var that = this;
  var connectReady = function (clusterName) {
    this._clusterMap[clusterName].connected = true;
    var allConnected = true;
    for (var i = 0; i < this._clusterNames.length; i++) {
      if (!this._clusterMap[this._clusterNames[i]].connected) {
        allConnected = false;
        break;
      }
    }
    if (allConnected) {
      this.removeAllListeners('client_connect', connectReady);
      this.emit('connect');
    }
  };

  that.on('client_connect', connectReady.bind(that));

  // sync config from diamond
  that._diamond = Diamond.create(that._diamondOptions);
  var dataGroup = that._dataId + '-GROUP';
  // get real cluster data group
  that._diamond.on(that._dataId, dataGroup, function (content) {
    debug('got %s:%s content: %j, local ip: %s', that._dataId, dataGroup, content, that._localIP);
    content = content.trim();
    // 1=ldbcommon-daily-1
    // 10.235.=ldbcommon-daily-1
    // 42.120=ldbcommon-daily-2
    var lines = content.split('\n');
    var closestGroup = null;
    var groups = [];
    for (var i = 0; i < lines.length; i++) {
      var items = lines[i].trim().split('=');
      var ipMatch = items[0] && items[0].trim();
      var name = items[1] && items[1].trim();
      if (items.length !== 2 || !ipMatch || !name) {
        continue;
      }
      if (!closestGroup && that._localIP.indexOf(ipMatch) === 0) {
        closestGroup = name;
      }
      if (groups.indexOf(name) < 0) {
        groups.push(name);
      }
    }
    if (!closestGroup) {
      // If can not found any match group, just use first group.
      closestGroup = groups[0];
    }

    var closestGroupIndex = groups.indexOf(closestGroup);
    var sortGroups = [closestGroup];
    for (var i = 0; i < groups.length; i++) {
      if (i === closestGroupIndex) {
        continue;
      }
      sortGroups.push(groups[i]);
    }
    that._clusterNames = sortGroups;

    debug('got closest %d:%j for %s, total %d groups: %j',
      closestGroupIndex, closestGroup, that._localIP, sortGroups.length, sortGroups);

    sortGroups.forEach(function (clusterName, clusterIndex) {
      // try to update group info
      if (!that._clusterMap[clusterName]) {
        that._clusterMap[clusterName] = {
          connected: false,
          client: null,
          readWeight: 0,
          writeWeight: 0
        };
        // remove old listeners
        that._diamond.removeAllListeners(that._dataId, clusterName);
        that._diamond.on(that._dataId, clusterName, that._handleClusterConfig.bind(that, clusterName, clusterIndex));
      }
    });
  }).onError('syncConfigError', function (err) {
    this.logger.warn('[%s] [%s] diamond sync config error: %s', Date(), process.pid, err);
  }.bind(this));
};

ClusterClient.prototype._handleClusterConfig = function (clusterName, clusterIndex, content) {
  var that = this;
  var config = null;
  try {
    config = JSON.parse(content);
  } catch (e) {
    try {
      // TODO: eval unsafe content will be danger
      config = eval('(' + content + ')');
    } catch (e) {
      this.logger.warn('[%s] [%s] JSON parse diamond config error: %s. content: %s', Date(), process.pid, e, content);
    }
  }
  debug('got cluster:%d `%s` config: %s, json decode: %j', clusterIndex, clusterName, content, config);
  // {"clusters":[{"address":{"master":"10.0.1.123:5198","slave":"10.1.1.456:5198","group":"group_2"},"readWeight":"1","writeWeight":"1"}]}
  var clusters = config && config.clusters || [];
  if (clusters.length === 0) {
    return;
  }

  // TODO: need to add all groups
  var needCluster = null;

  for (var i = 0; i < clusters.length; i++) {
    var c = clusters[i];
    var writeWeight = Number(c.writeWeight) || 0;
    if (!writeWeight) {
      continue;
    }

    if (!needCluster) {
      needCluster = c;
      continue;
    }

    var preWriteWeight = Number(needCluster.writeWeight) || 0;
    if (writeWeight > preWriteWeight) {
      // read weight bigger than prev cluster
      needCluster = c;
    }
  }

  if (!needCluster) {
    return;
  }

  var options = needCluster;
  var address = options.address;
  var master = address.master || '';
  master = master.split(':');

  if (!address.group || master.length !== 2) {
    return;
  }

  var groupName = address.group;
  var servers = [
    {host: master[0], port: parseInt(master[1], 10) || 5198}
  ];
  var slave = (address.slave || '').split(':');
  if (slave.length === 2) {
    servers.push({host: slave[0], port: parseInt(slave[1], 10) || 5198});
  }

  var cluster = this._clusterMap[clusterName];
  if (!cluster) {
    return;
  }

  if (!cluster.client) {
    cluster.client = new Client({
      namespace: this._namespace,
      timeout: this._timeout,
      heartbeatInterval: this._heartbeatInterval,
      syncInterval: this._syncInterval,
      clusterName: this._dataId + ':' + clusterIndex + ':' + clusterName,
      servers: servers,
      group: groupName,
      timer: this._timer,
      debug: this._debug,
      withHeader: this._withHeader,
    });
    cluster.client.once('connect', this.emit.bind(this, 'client_connect', clusterName));
    this._clusterMap[clusterName] = cluster;
  } else {
    cluster.client.updateGroup(groupName, servers);
  }
  cluster.readWeight = Number(options.readWeight) || 0;
  cluster.writeWeight = Number(options.writeWeight) || 0;
};

ClusterClient.prototype.close = function () {
  if (this._diamond) {
    this._diamond.close();
  }

  for (var k in this._clusterMap) {
    var cluster = this._clusterMap[k];
    cluster.client.close();
  }
  this._clusterMap = {};
  setTimeout(function () {
    debug('TairCluster %s closed.', this._dataId);
    this.emit('close');
  }.bind(this), 500);

  return this;
};

ClusterClient.prototype._getClients = function (isRead) {
  var clients = [];
  var field = isRead ? 'readWeight' : 'writeWeight';
  for (var i = 0; i < this._clusterNames.length; i++) {
    var cluster = this._clusterMap[this._clusterNames[i]];
    // TODO: 按权重来返回group列表
    // read: 优先最近的权重最高的
    // write: 权重大于0的都得写
    if (cluster[field] > 0 && cluster.connected) {
      clients.push(cluster.client);
    }
  }
  return clients;
};

var reads = ['get', 'prefixGet', 'prefixGets', 'getRange', 'getRangeOnlyValue', 'getRangeOnlyKey'];
reads.forEach(function (method) {
  ClusterClient.prototype[method] = function () {
    var lastIndex = arguments.length - 1;
    var args = Array.prototype.slice.call(arguments, 0, lastIndex);
    var callback = arguments[lastIndex];
    var clients = this._getClients(true);

    if (clients.length === 0) {
      var err = new Error('No readable clusters');
      err.name = 'TairNoReadClusterError';
      return callback(err);
    }

    var index = 0;
    var lastError = null;
    var next = function () {
      if (index >= clients.length) {
        return callback(lastError);
      }

      var client = clients[index++];
      args[lastIndex] = function (err, data, res) {
        if (err && err.name === 'TairDataNodeNotAliveError') {
          lastError = err;
          return next();
        }
        callback(err, data, res);
      };
      client[method].apply(client, args);
    };
    next();

    return this;
  };
});

var writes = ['put', 'prefixPuts', 'prefixPut', 'incr', 'decr', 'setCount', 'invalid', 'lock', 'unlock'];
writes.forEach(function (method) {
  ClusterClient.prototype[method] = function () {
    var lastIndex = arguments.length - 1;
    var callback = arguments[lastIndex];
    if (typeof callback !== 'function') {
      // keep silent when callback not affort
      callback = utility.noop;
      lastIndex++;
    }
    var args = Array.prototype.slice.call(arguments, 0, lastIndex);
    var clients = this._getClients(false);
    if (clients.length === 0) {
      var err = new Error('No writable clusters');
      err.name = 'TairNoWriteClusterError';
      return callback(err);
    }

    var ep = eventproxy.create();
    ep.on('error', function (err, data, res) {
      ep.unbind();
      callback(err, data, res);
    });
    clients.forEach(function (client) {
      args[lastIndex] = function (err, data, res) {
        if (err) {
          return ep.emit('error', err, data, res);
        }

        ep.emit('done', [data, res]);
      };
      client[method].apply(client, args);
    });
    ep.after('done', clients.length, function (results) {
      var result = results[0];
      callback(null, result[0], result[1]);
    });

    return this;
  };
});
ClusterClient.prototype.set = ClusterClient.prototype.put;

