/*!
 * node-tair - lib/connection.js
 *
 * Copyright(c) 2013 fengmk2 <fengmk2@gmail.com> (http://fengmk2.github.com)
 * MIT Licensed
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:connection');
var net = require('net');
var EventEmitter = require('events').EventEmitter;
var util = require('util');
var Readable = require('readable-stream').Readable;
var utility = require('utility');
var cons = require('./const');
var ByteBuffer = require('byte');

var MAX_INT32 = 0x7FFFFFFF; // 2147483647;
var REQUEST_ID = 0;

var CONNECTION_ID = 0;

function Connection(options) {
  EventEmitter.call(this);

  this.id = ++CONNECTION_ID;
  this.host = options.host;
  this.port = options.port;
  this.name = 'conn#' + this.id + ' (' + this.host + ':' + this.port + ')';
  this.timeout = options.timeout || 5000;
  this._debug = options.debug;
  this._timer = options.timer || Date;
  this.calls = {};
  this._lastError = null;

  debug('%s connecting...', this.name);
  this.socket = net.connect({
    host: this.host,
    port: this.port
  });
  this.socket.setNoDelay(true);
  this.startTime = Date.now();
  this.out = this.socket;
  this.io = this.socket;
  if (typeof this.io.read !== 'function') {
    this.io = new Readable().wrap(this.socket);
    this.io.on('error', utility.noop);
  }

  this._reading = 'header';
  this.io.on('readable', this._handleResponse.bind(this));

  this.socket.once('timeout', this._handleTimeout.bind(this));
  // The 'close' event will be called directly following 'error' event.
  this.socket.once('error', this._handleError.bind(this));
  this.socket.once('close', this._handleClose.bind(this));
  this.socket.once('connect', function () {
    debug('%s connected', this.name);
    this.emit('connect');
  }.bind(this));
}

util.inherits(Connection, EventEmitter);

Connection.prototype._handleTimeout = function () {
  var err = new Error('The socket timeout event emit. ' + this.name);
  err.name = 'TairConnectionTimeoutError';
  this._lastError = err;
  this.socket.destroy();
};

Connection.prototype._handleError = function (err) {
  this._lastError = err;
};

Connection.prototype._handleClose = function (hadError) {
  this.emit('close'); // tell user not use this connection first.
  var err = this._lastError;
  if (!err) {
    if (hadError) {
      err = new Error('The socket was closed due to a transmission error. ' + this.name);
      err.name = 'TairConnectionClosedByTransmissionError';
    } else {
      err = new Error('The socket was closed. ' + this.name);
      err.name = 'TairConnectionClosedError';
    }
  }
  if (err.name === 'Error') {
    if (err.message === 'write EPIPE') {
      err.name = 'TairConnectionWasClosedError';
    } else {
      err.name = 'TairConnectionUnknowError';
    }
    err.message = err.message + ' ' + this.name;
  }
  debug('%s was closed, error: %s', this.name, err);
  this._cleanupCalls(err);
};

Connection.prototype._cleanupCalls = function (err) {
  var calls = this.calls;
  this.calls = {};
  for (var id in calls) {
    var req = calls[id];
    req.setError(err);
  }
};

Connection.prototype.close = function () {
  this.socket.end();
};

Connection.prototype.call = function (packet, timeout, callback) {
  if (typeof timeout === 'function') {
    callback = timeout;
    timeout = null;
  }
  timeout = timeout || this.timeout;
  var req = new Request(this.name, packet, timeout, {debug: this._debug, timer: this._timer}, callback);
  req.encode(this.out);
  this.calls[req.id] = req;
  var that = this;
  req.on('finish', function (err, res) {
    delete that.calls[req.id];
    callback(err, res);
  });
  return req;
};

Connection.prototype._handleResponse = function () {
  if (this._reading === 'header') {
    return this._handleResponseHeader();
  }
  this._handleResponseBody();
};

Connection.prototype._handleResponseHeader = function () {
  // flag, req id, packet code, body length = 16 bytes
  var header = this.io.read(16);
  if (!header) {
    return;
  }

  if (header.length !== 16) {
    var err = new Error('Tair packet header wrong size: ' + header.length + ' bytes. ' + this.name);
    err.name = 'TairWrongHeaderSizeError';
    this._handleError(err);
    return this.socket.destroy();
  }

  var buf = ByteBuffer.wrap(header);
  var flag = buf.getInt();
  if (flag !== cons.TAIR_PACKET_FLAG) {
    var err = new Error('Tair packet flag wrong, flag: ' + flag + '. ' + this.name);
    err.name = 'TairWrongPacketFlagError';
    this._handleError(err);
    return this.socket.destroy();
  }

  var reqId = buf.getInt();
  var packetCode = buf.getInt();
  var length = buf.getInt();
  var req = this.calls[reqId];
  if (req) {
    req.setResponseInfo(packetCode, length);
  }

  debug('%s req#%d got response header, length: %d, packetCode: %d, body length: %d',
    this.name, reqId, header.length, packetCode, length);

  this._reading = 'body';
  this._reponseRequestId = reqId;
  this._reponseBodyLength = length;
  this._handleResponseBody();
};

Connection.prototype._handleResponseBody = function () {
  var reqId = this._reponseRequestId;
  var length = this._reponseBodyLength;
  var body = this.io.read(length);
  debug('%s req#%d try to receive %d bytes, got %d bytes',
    this.name, reqId, length, body && body.length || 0);
  if (!body) {
    return;
  }

  if (body.length !== length) {
    var err = new Error('Tair packet body wrong size: ' + body.length + ' bytes. ' + this.name);
    err.name = 'TairWrongBodySizeError';
    this._handleError(err);
    return this.socket.destroy();
  }

  var req = this.calls[reqId];
  if (req) {
    // req timeout event will cause req remove before response
    req.decode(body);
  }

  debug('%s req#%d done, next()', this.name, reqId);
  this._reading = 'header';
  this._handleResponseHeader();
};

function Request(connectionName, param, timeout, options, callback) {
  EventEmitter.call(this);
  options = options || {};

  this.id = this.getId();
  this.connectionName = connectionName;
  this.param = param;
  this.timeout = timeout;
  this.error = null;
  this.finished = false;
  this._debug = options.debug;
  this._nowTimer = options.timer || Date;
  this.startTime = this._nowTimer.now();
  this.res = {
    reqId: this.id,
    reqSize: 0,
    packetCode: 0,
    bodySize: 0,
    body: null,
    use: 0
  };
  this.timer = setTimeout(this._handleTimeout.bind(this), timeout);
}

util.inherits(Request, EventEmitter);

Request.prototype.getId = function () {
  REQUEST_ID++;
  if (REQUEST_ID >= MAX_INT32) {
    REQUEST_ID = 1;
  }
  return REQUEST_ID;
};

Request.prototype.encode = function (out) {
  var buf = this.param.encode(this.id);
  out.write(buf);
  this.res.reqSize = buf.length;
  debug('%s req#%d send %s packet, %d bytes',
    this.connectionName, this.id, this.param.constructor.name, buf.length);
};

Request.prototype._finish = function () {
  if (this.timer) {
    clearTimeout(this.timer);
    this.timer = null;
  }
  // TODO: why we need this?
  // if (this.finished) {
  //   return;
  // }
  this.finished = true;
  this.res.use = this._nowTimer.now() - this.startTime;
  this.emit('finish', this.error, this.res);
};

Request.prototype.setResponseInfo = function (packetCode, bodySize) {
  this.res.packetCode = packetCode;
  this.res.bodySize = bodySize;
};

Request.prototype.decode = function (data) {
  var buf = ByteBuffer.wrap(data);
  var that = this;
  if (that._debug) {
    that.res._body = data;
  }
  if (typeof that.param.asyncDecode === 'function') {
    that.param.asyncDecode(buf, function (err, body) {
      if (err) {
        return that.setError(err);
      }

      that.res.body = body;
      that._finish();
    });
  } else {
    try {
      that.res.body = that.param.decode(buf);
    } catch (e) {
      buf.position(data.length);
      e.name = 'Tair' + that.param.constructor.name + 'Decode' + e.name;
      e.data = {
        connectionName: that.connectionName,
        reqId: that.id,
        resBuffer: buf.toString()
      };
      return that.setError(e);
    }

    that._finish();
  }
};

Request.prototype.setError = function (err) {
  this.error = err;
  this._finish();
};

Request.prototype._handleTimeout = function () {
  var err = new Error('Tair server no response in ' + this.timeout + ' ms, ' + this.connectionName + ' req#' + this.id);
  err.name = 'Tair' + this.param.constructor.name + 'RequestTimeoutError';
  this.setError(err);
};

Connection.Request = Request;

module.exports = Connection;
