/*!
 * node-tiar - lib/packets/get.js
 *
 * Copyright(c) 2012 - 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 *   sunfang1cn <kate.sf@taobao.com>
 */

"use strict";

/**
 * Module dependencies.
 */

var util = require('util');
var ByteBuffer = require('byte');
var cons = require('../const');
var BasePacket = require('./base');
var MixedKey = require('./mixed_key');

/**
 * Get packet
 *
 * @param {String|Buffer|Number} key
 * @param {Transcoder} transcoder
 * @param {Object} [options]
 *  - {Number} [namespace] default is 0
 */
function Get(key, transcoder, options) {
  options = options || {};
  BasePacket.call(this, cons.TAIR_REQ_GET_PACKET);

  this.key = key;

  this.namespace = options.namespace || 0;
  this.transcoder = transcoder;

  this.keyBytes = this.transcoder.encode(this.key);
  this.keys = [];

  this.addKey(this.keyBytes);
}

util.inherits(Get, BasePacket);

Get.ResponseCode = cons.TAIR_RESP_GET_PACKET;

Get.prototype.addKey = function (key) {
  this.keys.push(key);
};

Get.prototype.encode = function (reqId) {
  var keyBytes = this.keyBytes;
  // total length: 1 + 2 + 4 + 36 + key (String length: 4 + bytes.length) = 47 + key bytes length
  // format: | 0 (1 byte, body start) | namespace (2 bytes)
  //         | meta 36 bytes
  //         | key count (1, 4 bytes)
  //         | key1 meta flags (36 bytes)
  //         | key1 length (4 bytes) | key1 bytes

  var keyList = [];
  var bodyLength = 7;
  var isMixedKey = this.key instanceof MixedKey;
  for (var i = 0; i < this.keys.length; i++) {
    var key = this.keys[i];
    var keyBytes = Buffer.isBuffer(key) ? key : this.transcoder.encode(key);
    keyList.push(keyBytes);
    bodyLength += 40 + keyBytes.length;
  }

  var totalLength = cons.PACKET_HEADER_SIZE + 4 + bodyLength;
  var bytes = ByteBuffer.allocate(totalLength);
  // request packet format: | header (12 bytes) | body |
  // header: | flag: 6d 42 64 54 | request id (4 bytes) | packet code (4 bytes) |
  bytes.putInt(cons.TAIR_PACKET_FLAG); // packet flag
  bytes.putInt(reqId); // request id
  // | packet code (4 bytes) | body length (4 bytes) | body |
  bytes.putInt(this.code); // 12

  bytes.putInt(bodyLength);

  // body
  bytes.put(0);
  bytes.putShort(this.namespace);
  bytes.putInt(keyList.length);

  for (var i = 0; i < keyList.length; i++) {
    var keyBytes = keyList[i];
    this.fillMetaFlag(bytes);
    var len = keyBytes.length;
    if (isMixedKey) {
      var prefixSize = this.key.prefixSize;
      len = prefixSize;
      len <<= 22;
      len |= keyBytes.length;
    }
    bytes.putInt(len);
    bytes.put(keyBytes);
  }
  return bytes.array();
};

Get.unpack = function (body, transcoder) {
  var bytes = ByteBuffer.wrap(body);
  var data = {
    namespace: 0,
    key: null,
  };
  bytes.skip(1); // body padding
  data.namespace = bytes.getShort();
  var keyCount = bytes.getInt();
  if (keyCount === 1) {
    // skip meta
    bytes.skip(36);
    var msize = bytes.getInt();
    var size = msize & 0x3FFFFF;
    var prefixSize = msize >> 22;
    if (size > 0) {
      data.key = transcoder.decode(bytes.read(size), 0, size, prefixSize);
    }
  } else if (keyCount > 1) {
    data.keys = [];
    for (var i = 0; i < keyCount; i++) {
      // skip meta
      bytes.skip(36);
      var msize = bytes.getInt();
      var size = msize & 0x3FFFFF;
      var prefixSize = msize >> 22;
      if (size > 0) {
        var key = transcoder.decode(bytes.read(size), 0, size, prefixSize);
        if (i === 0) {
          data.key = key;
        }
        data.keys.push(key);
      }
    }
  }
  return data;
};

Get.prototype.decode = function (buf) {
  var res = {
    version: 0,
    code: 0,
    count: 0,
    key: null,
    value: null,
    meta: {
      isMerged: false,
      area: 0,
      serverFlag: 0,
    }
  };

  res.version = buf.getInt();
  res.code = buf.getInt();
  res.count = buf.getInt();

  if (res.count === 0) {
    if (res.code === -3998) {
      res.code = 1;
    }
    // -3988: data expired
    if (res.code === -3988) {
      res.code = 3988;
    }
    return res;
  }

  buf.skip(36);
  // meta.isMerged = buf.get(); // 1
  // meta.area = buf.getInt(); // 5
  // meta.serverFlag = buf.getShort(); // 7
  // meta.magic = buf.getShort(); // 9
  // meta.checkSum = buf.getShort(); // 11
  // meta.keySize = buf.getShort(); // 13
  // meta.version = buf.getShort(); // 15
  // meta.padSize = buf.getInt(); // 19
  // meta.valueSize = buf.getInt(); // 23
  // meta.flag = buf.get(); // 24
  // meta.cdate = buf.getInt(); // 28
  // meta.mdate = buf.getInt(); // 32
  // meta.edate = buf.getInt(); // 36

  var msize = buf.getInt();
  var size = msize & 0x3FFFFF;
  var prefixSize = msize >> 22;

  if (size > 0) {
    res.key = this.transcoder.decode(buf.read(size), 0, size, prefixSize);
  }

  // skip 36 bytes meta flag
  buf.skip(36);
  var valueLength = buf.getInt();
  if (valueLength > 0) {
    res.value = this.transcoder.decode(buf.read(valueLength));
  }
  return res;
};

Get.pack = function (res, transcoder) {
  var count = 1;
  var size;
  var valueBytes;
  var keyBytes;
  if (!res.value) {
    count = 0;
    size = 12;
  } else {
    valueBytes = transcoder.encode(res.value);
    var ksize = 0;
    if (res.key) {
      keyBytes = transcoder.encode(res.key);
      ksize = keyBytes.length;
    }
    // size = 12 + 36 + 4 + keyBytes.length + 36 + 4 + valueBytes.length;
    size = 92 + ksize + valueBytes.length;
  }
  var bytes = ByteBuffer.allocate(size);
  bytes.putInt(res.version || 0);
  bytes.putInt(res.code || 0);
  bytes.putInt(count);

  if (count === 0) {
    return bytes.array();
  }

  // fill meta
  bytes.put(cons.EMPTY_META);
  if (keyBytes) {
    bytes.putInt(keyBytes.length);
    bytes.put(keyBytes);
  } else {
    bytes.putInt(0);
  }

  bytes.put(cons.EMPTY_META);
  bytes.putInt(valueBytes.length);
  bytes.put(valueBytes);
  return bytes.array();
};

Get.packError = function (res, transcoder) {
  var bytes = ByteBuffer.allocate(12);
  bytes.putInt(res.version);
  bytes.putInt(res.value);
  bytes.putInt(0);
  return bytes.array();
};

module.exports = Get;
