/*!
 * node-tair - lib/packets/get_group.js
 *
 * Copyright(c) 2013 fengmk2 <fengmk2@gmail.com> (http://fengmk2.github.com)
 * MIT Licensed
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:packets:get_group');
var utility = require('utility');
var util = require('util');
var zlib = require('zlib');
var ByteBuffer = require('byte');
var cons = require('../const');
var BasePacket = require('./base');
var ByteBuffer = require('byte');
var longToHost = require('../utils').longToHost;
var hostToLong = require('../utils').hostToLong;

var Caches = {};
var CacheCount = 0;

function GetGroup(version, name, clusterName) {
  BasePacket.call(this, cons.TAIR_REQ_GET_GROUP_NEW_PACKET);

  this.version = version;
  this.name = name;
  this._cacheKey = clusterName + ':' + name + ':' + version;
  this.nameBytes = Buffer.isBuffer(name) ? name : new Buffer(name);
}

util.inherits(GetGroup, BasePacket);

GetGroup.ResponseCode = cons.TAIR_RESP_GET_GROUP_NEW_PACKET;

GetGroup.prototype.encode = function (reqId) {
  var bodyLength = 4 + 4 + this.nameBytes.length + 1;
  var totalLength = cons.PACKET_HEADER_SIZE + 4 + bodyLength;
  var bytes = ByteBuffer.allocate(totalLength);
  // request packet format: | header (12 bytes) | body |
  // header: | flag: 6d 42 64 54 (4 bytes) | request id (4 bytes) | packet code (4 bytes) |
  bytes.putInt(cons.TAIR_PACKET_FLAG); // packet flag
  bytes.putInt(reqId); // request id
  // | packet code (4 bytes) | body length (4 bytes) | body |
  bytes.putInt(this.code); // 12

  bytes.putInt(bodyLength);
  // | version (4 bytes) | name (CString, 4 bytes + bytes.length + 1) |
  bytes.putInt(this.version);

  // CString
  bytes.putInt(this.nameBytes.length + 1);
  bytes.put(this.nameBytes);
  bytes.put(0);
  return bytes.array();
};

GetGroup.unpack = function (body, transcoder) {
  var bytes = ByteBuffer.wrap(body);
  var data = {
    version: bytes.getInt(),
    name: bytes.getCString()
  };
  return data;
};

GetGroup.prototype.asyncDecode = function (buf, callback) {
  // | bucket (4 bytes) | copy (4 bytes) | version (4 bytes)
  // | config count (4 bytes) | name (CString) | value (CString) | ...
  // | server data length (4 Bytes) | server data (zlib inflate format, LITTLE_ENDIAN, ip:port)
  // | alive node count (4 bytes) | node 1 (port:ip 8 bytes) | node 2 | ... |
  var cacheKey = this._cacheKey;
  var cacheGroup = Caches[cacheKey];
  var _bytes = buf._bytes;
  var byteMD5 = utility.md5(_bytes);
  if (cacheGroup && cacheGroup.byteSize === _bytes.length && cacheGroup.byteMD5 === byteMD5) {
    cacheGroup.byteCache = true;
    debug('GetGroup: got %s from caches: %s, %d bytes, servers bucket %d, nodes: %j',
      cacheKey,
      byteMD5, cacheGroup.byteSize, cacheGroup.servers[0].length, cacheGroup.nodes);
    return callback(null, cacheGroup);
  }

  var res = {
    bucket: 0,
    copy: 0,
    version: 0,
    configCount: 0,
    config: {},
    servers: [
      []
    ],
    nodeCount: 0,
    nodes: {},
    invalidateServers: [],
    downServers: {},
    downServerCount: 0,
    byteSize: _bytes.length,
    byteMD5: byteMD5,
  };

  res.bucket = buf.getInt();
  res.copy = buf.getInt();
  res.version = buf.getInt();
  res.configCount = buf.getInt();
  for (var i = 0; i < res.configCount; i++) {
    var name = buf.getCString();
    var value = buf.getCString();
    res.config[name] = value;
  }
  if (res.config.invalidate_server) {
    // '10.235.144.193:5126,10.235.144.194:5126'
    var servers = res.config.invalidate_server.split(',');
    for (var i = 0; i < servers.length; i++) {
      var server = servers[i].trim().split(':');
      if (server.length !== 2) {
        continue;
      }
      res.invalidateServers.push({host: server[0], port: parseInt(server[1], 10)});
    }
  }
  if (res.config.tmp_down_server) {
    // down servers: 'ip1:port1;ip2:port2;...'
    var servers = res.config.tmp_down_server.split(';');
    for (var i = 0; i < servers.length; i++) {
      var server = servers[i].trim();
      if (server.split(':').length !== 2) {
        continue;
      }
      res.downServerCount++;
      res.downServers[server] = true;
    }
  }

  var serverDataLength = buf.getInt();
  var serverData = buf.read(serverDataLength);

  res.nodeCount = buf.getInt();
  for (var i = 0; i < res.nodeCount; i++) {
    var port = buf.getInt();
    var ip = buf.getInt();
    res.nodes[ip + ':' + port] = {host: longToHost(ip), port: port, ip: ip};
  }

  zlib.inflate(serverData, function (err, buffer) {
    // LITTLE_ENDIAN
    if (err) {
      return callback(err, res);
    }
    var data = ByteBuffer.wrap(buffer);
    data.order(ByteBuffer.LITTLE_ENDIAN);
    var bucketIndex = 0;
    var serverIndex = 0;
    while (data.position() < buffer.length) {
      var ip = data.getInt();
      var port = data.getInt();
      res.servers[serverIndex].push({host: ip, port: port});
      bucketIndex++;
      if (bucketIndex >= res.bucket) {
        bucketIndex = 0;
        serverIndex++;
        res.servers[serverIndex] = [];
      }
    }

    if (!Caches[cacheKey]) {
      CacheCount++;
      if (CacheCount > 5) {
        // if cache group info item bigger than 5, clean all of them.
        Caches = {};
        CacheCount = 1;
      }
    }
    Caches[cacheKey] = res;

    debug('GetGroup: %s servers bucket %d, %d nodes: %j, %d invalidateServers: %j, %d down servers: %j',
      cacheKey, res.servers[0].length,
      res.nodeCount, res.nodes,
      res.invalidateServers.length, res.invalidateServers,
      res.downServerCount, res.downServers);
    callback(null, res);
  });
};

GetGroup.asyncPack = function (res, transcoder, callback) {
  var bytes = ByteBuffer.allocate(500);
  bytes.putInt(res.bucket);
  bytes.putInt(res.copy);
  bytes.putInt(res.version);
  bytes.putInt(res.configCount);
  if (res.configCount > 0) {
    for (var k in res.config) {
      bytes.putCString(k);
      bytes.putCString(res.config[k]);
    }
  }

  var serverBytes = ByteBuffer.allocate(200);
  serverBytes.order(ByteBuffer.LITTLE_ENDIAN);
  for (var i = 0; i < res.servers.length; i++) {
    var server = res.servers[i];
    serverBytes.putInt(hostToLong(server.host));
    serverBytes.putInt(Number(server.port));
  }
  zlib.deflate(serverBytes.array(), function (err, buf) {
    if (err) {
      return callback(err);
    }
    bytes.putInt(buf.length);
    bytes.put(buf);

    bytes.putInt(res.servers.length);
    for (var i = 0; i < res.servers.length; i++) {
      var server = res.servers[i];
      bytes.putInt(Number(server.port));
      bytes.putInt(hostToLong(server.host));
    }

    callback(null, bytes.array());
  });
};

module.exports = GetGroup;
