/*!
 * node-tair - lib/packets/get_range.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var util = require('util');
var ByteBuffer = require('byte');
var BasePacket = require('./base');
var cons = require('../const');

/**
 * Prefix get range packet
 *
 * @param {MixedKey} startKey
 * @param {MixedKey} endKey
 * @param {Number} cmdType
 * @param {Transcoder} transcoder
 * @param {Object} options
 *  - {Number} [namespace] default is 0
 */
function GetRange(startKey, endKey, cmdType, offset, limit, transcoder, options) {
  BasePacket.call(this, cons.TAIR_REQ_GET_RANGE_PACKET, transcoder);

  this.key = startKey;
  this.endKey = endKey;
  this.cmdType = cmdType;
  this.offset = offset;
  this.limit = limit;
  this.namespace = options.namespace || 0;

  this.keyBytes = this.transcoder.encode(this.key);
  this.endKeyBytes = this.transcoder.encode(this.endKey);
  this.prefixSize = this.key.prefixSize;
}

util.inherits(GetRange, BasePacket);

GetRange.prototype.encode = function (reqId) {
  // format: | 0 (1 byte, body start) | cmdType (2 bytes) | namespace (2 bytes) => 5 bytes
  //         | offset (4 bytes) | limit (4 bytes)                               => 8 bytes
  //         | start key meta (36 bytes)
  //         | start key length (4 bytes)
  //         | start key bytes
  //         | end key meta (36 bytes)
  //         | end key length (4 bytes)
  //         | end key bytes
  var bodyLength = 93 + this.keyBytes.length + this.endKeyBytes.length;

  var totalLength = cons.PACKET_HEADER_SIZE + 4 + bodyLength;
  var bytes = ByteBuffer.allocate(totalLength);
  // request packet format: | header (12 bytes) | body |
  // header: | flag: 6d 42 64 54 | request id (4 bytes) | packet code (4 bytes) |
  bytes.putInt(cons.TAIR_PACKET_FLAG); // packet flag
  bytes.putInt(reqId); // request id
  // | packet code (4 bytes) | body length (4 bytes) | body |
  bytes.putInt(this.code); // 12

  bytes.putInt(bodyLength);

  // body
  bytes.put(0);
  bytes.putShort(this.cmdType);
  bytes.putShort(this.namespace);
  bytes.putInt(this.offset);
  bytes.putInt(this.limit); // 13

  // start key
  this.fillMetaFlag(bytes);
  bytes.putInt((this.prefixSize << cons.PREFIX_KEY_OFFSET) | this.keyBytes.length);
  bytes.put(this.keyBytes);

  // end key
  this.fillMetaFlag(bytes);
  bytes.putInt((this.prefixSize << cons.PREFIX_KEY_OFFSET) | this.endKeyBytes.length);
  bytes.put(this.endKeyBytes);

  return bytes.array();
};

GetRange.prototype.decode = function (buf) {
  var res = {
    version: 0,
    code: 0,
    // originalCode: 0,
    items: [],
    count: 0
  };

  res.version = buf.getInt();
  res.code = buf.getInt();
  var cmd = buf.getShort();
  var count = buf.getInt();
  var flag = buf.getShort();

  if (cmd === cons.CMD_RANGE_ALL || cmd === cons.CMD_RANGE_ALL_REVERSE) {
    count /= 2;
  }

  res.count = count;

  for (var i = 0; i < count; i++) {
    var item = {key: null, value: null};
    if (cmd === cons.CMD_RANGE_ALL || cmd === cons.CMD_RANGE_ALL_REVERSE ||
        cmd === cons.CMD_RANGE_KEY_ONLY || cmd === cons.CMD_RANGE_KEY_ONLY_REVERSE ||
        cmd === cons.CMD_DEL_RANGE || cmd === cons.CMD_DEL_RANGE_REVERSE) {
      buf.skip(36);

      var msize = buf.getInt();
      size = msize & 0x3FFFFF;
      var prefixSize = msize >> 22;

      if (size > 0) {
        item.key = this.transcoder.decode(buf.read(size), 0, size, prefixSize);
      }
    }

    if (cmd === cons.CMD_RANGE_ALL || cmd === cons.CMD_RANGE_VALUE_ONLY ||
        cmd === cons.CMD_RANGE_ALL_REVERSE || cmd === cons.CMD_RANGE_VALUE_ONLY_REVERSE) {
      buf.skip(36);
      // removeMetas();
      // if (cmd == TairConstant.CMD_RANGE_VALUE_ONLY){
      //   de.decodeMeta(byteBuffer);
      // } else {
      //   new DataEntry().decodeMeta(byteBuffer);
      // }
      // data
      var size = buf.getInt();
      if (size > 0) {
        item.value = this.transcoder.decode(buf.read(size), 0, size);
      }
    }

    res.items.push(item);
  }

  if (count > 1) {
    // TODO: what is this ?
    // var pc = buf.getInt();
    // if (pc > 0) {
    //   for (var i = 0; i < pc; i++) {
    //     buf.skip(36);
    //     var size = buf.getInt();
    //     if (size > 0) {
    //       var v = this.transcoder.decode(buf.read(size), 0, size);
    //       console.log(pc, i, v);
    //     }
    //   }
    // }
  }

  return res;
};

module.exports = GetRange;
