/*!
 * node-tair - lib/packets/invalid.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:packets:invalid');
var util = require('util');
var zlib = require('zlib');
var ByteBuffer = require('byte');
var cons = require('../const');
var BasePacket = require('./base');

/**
 * Invalid packet
 *
 * @param {String|Buffer} key
 * @param {String} groupName
 * @param {Transcoder} transcoder
 * @param {Object} [options]
 *  - {Number} [namespace] default is 0
 */
function Invalid(key, groupName, transcoder, options) {
  BasePacket.call(this, cons.TAIR_REQ_INVALID_PACKET);

  this.key = key;
  this.transcoder = transcoder;
  this.groupName = groupName;

  options = options || {};
  this.namespace = options.namespace || 0;
  this.isSync = 0;

  this.keyBytes = this.transcoder.encode(this.key);
}

util.inherits(Invalid, BasePacket);

Invalid.prototype.encode = function (reqId) {
  var keyBytes = this.keyBytes;
  var groupNameBytes = new Buffer(this.groupName);

  // total length: 1 + 2 + 4 + 36 + key (String length: 4 + bytes.length) + groupName(CString) + 4
  //                = 43 + 4 + key bytes length + 5 + groupName bytes length + 4
  //                = 56 + key bytes length + groupName bytes length
  // format: | 0 (1 byte, body start) | namespace (2 bytes)
  //         | key count (1, 4 bytes)
  //         | key1 meta flags (36 bytes)
  //         | key1 String
  //         | groupName CString | isSync (4 bytes)
  var bodyLength = 56 + keyBytes.length + groupNameBytes.length;

  var totalLength = cons.PACKET_HEADER_SIZE + 4 + bodyLength;
  var bytes = ByteBuffer.allocate(totalLength);
  // request packet format: | header (12 bytes) | body |
  // header: | flag: 6d 42 64 54 | request id (4 bytes) | packet code (4 bytes) |
  bytes.putInt(cons.TAIR_PACKET_FLAG); // packet flag
  bytes.putInt(reqId); // request id
  // | packet code (4 bytes) | body length (4 bytes) | body |
  bytes.putInt(this.code); // 12

  bytes.putInt(bodyLength);

  // body
  bytes.put(0);
  bytes.putShort(this.namespace);
  bytes.putInt(1);

  this.fillMetaFlag(bytes);
  bytes.putInt(keyBytes.length);
  bytes.put(keyBytes);

  // CString
  bytes.putInt(groupNameBytes.length + 1);
  bytes.put(groupNameBytes);
  bytes.put(0);

  bytes.putInt(this.isSync);
  debug('encode namespace: %d key: %s (%d), group: %s (%d), body length: %d',
    this.namespace, this.key, keyBytes.length, this.groupName, groupNameBytes.length, bodyLength);

  return bytes.array();
};

module.exports = Invalid;
