/*!
 * node-tair - lib/packets/prefix_puts.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:packets:prefix_puts');
var util = require('util');
var ByteBuffer = require('byte');
var cons = require('../const');
var BasePacket = require('./base');
var MixedKey = require('./mixed_key');

/**
 * Prefix Puts packet
 *
 * @param {String|Buffer|Number} pkey
 * @param {Array} keyValues
 *  - {Array} kv `[key, value, version, expire]`, optional: version, expire
 * @param {Transcoder} transcoder
 * @param {Object} options
 *  - {Number} [namespace] defautl is 0
 *  - {Number} [expired] expired time, default is 0, never expired.
 *  - {Number} [version] default is 0
 */
function PrefixPuts(pkey, keyValues, transcoder, options) {
  BasePacket.call(this, cons.TAIR_REQ_PREFIX_PUTS_PACKET, false);

  this.pkey = pkey;
  this.keyValues = keyValues;

  options = options || {};
  this.namespace = options.namespace || 0;
  this.expired = options.expired || 0;
  this.version = options.version || 0;
  this.transcoder = transcoder;

  var skey = keyValues[0][0];
  this.key = new MixedKey(this.transcoder, pkey, skey);
  this.keyBytes = this.transcoder.encode(this.key);
}

util.inherits(PrefixPuts, BasePacket);

PrefixPuts.prototype.encode = function (reqId) {
  var keyValueBytesList = [];
  var keyValueBytesSize = 0;

  for (var i = 0; i < this.keyValues.length; i++) {
    // [key, value, version, expire, 4:prefixSize, 5:flag]
    var kv = this.keyValues[i];
    var mixedKey = new MixedKey(this.transcoder, this.pkey, kv[0]);
    var kBytes = this.transcoder.encode(mixedKey);
    var vBytes = this.transcoder.encode(kv[1]);
    keyValueBytesList.push([kBytes, vBytes]);
    // TODO: support IncData
    kv[4] = mixedKey.prefixSize;
    kv[5] = 0;
    keyValueBytesSize += kBytes.length + vBytes.length;
    keyValueBytesSize += 40 * 2; //~ meta + length for each key&value
  }

  var pkeyBytes = this.transcoder.encode(this.pkey);

  // total length: 1 + 2 + 36 + pkey (String length: 4 + bytes.length) + 4 + keyValues
  //                = 47 + keyValues
  // format: | 0 (1 byte, body start) | namespace (2 bytes)
  //         | pkey meta flags (36 bytes)
  //         | pkey length (4 bytes, CString) | pkey bytes
  //         | keyValue length (4 bytes)
  //         | keyValues
  var bodyLength = 47 + pkeyBytes.length + keyValueBytesSize;
  // out.writeInt32(bodyLength); // 16

  var totalLength = cons.PACKET_HEADER_SIZE + 4 + bodyLength;
  var bytes = ByteBuffer.allocate(totalLength);
  // request packet format: | header (12 bytes) | body |
  // header: | flag: 6d 42 64 54 | request id (4 bytes) | packet code (4 bytes) |
  bytes.putInt(cons.TAIR_PACKET_FLAG); // packet flag
  bytes.putInt(reqId); // request id
  // | packet code (4 bytes) | body length (4 bytes) | body |
  bytes.putInt(this.code); // 12

  bytes.putInt(bodyLength);

  // body
  bytes.put(0); // 17
  bytes.putShort(this.namespace); // 19

  // pkey
  this.fillMetaFlag(bytes); // 19 + 36 = 55
  bytes.putInt(pkeyBytes.length);
  bytes.put(pkeyBytes); //

  //~ encode key/value
  bytes.putInt(keyValueBytesList.length); // 87

  for (var i = 0; i < keyValueBytesList.length; i++) {
    var item = keyValueBytesList[i];
    // [0:key, 1:value, 2:version, 3:expire, 4:prefixSize, 5:flag]
    var kv = this.keyValues[i];
    var version = kv[2] || 0;
    var expired = kv[3] || 0;
    var prefixSize = kv[4] || 0;
    var flag = kv[5] || 0;

    //~ key
    // var meta = new Buffer(36);
    // meta.fill(0);
    // meta.writeInt16BE(version, 13);
    // meta.writeInt32BE(expired, 32);
    // out.write(meta);
    bytes.put(cons.META_FLAG); // 7
    bytes.putShort(0); // magic 9
    bytes.putShort(0); // checkSum 11
    bytes.putShort(0); // keySize 13
    bytes.putShort(version); // version 15
    bytes.putInt(0); // padSize 19
    bytes.putInt(0); // valueSize 23
    bytes.put(0); // flag 24
    bytes.putInt(0); // cdate 28
    bytes.putInt(0); // mdate 32
    bytes.putInt(expired); // edate 36 => 87 + 36 = 123

    var keyBytes = item[0];
    var len = (prefixSize << 22) | keyBytes.length;
    bytes.putInt(len);
    bytes.put(keyBytes);

    //~ value
    var valueBytes = item[1];
    this.fillMetaFlag(bytes, flag);
    bytes.putInt(valueBytes.length);
    bytes.put(valueBytes);
  }

  return bytes.array();
};

PrefixPuts.prototype.decode = function (buf) {
  var res = {
    version: buf.getInt(),
    code: buf.getInt(),
    message: buf.getString(),
    count: buf.getInt(),
    codes: null
  };
  if (res.count > 0) {
    var count = res.count;
    var codes = {};
    res.codes = codes;
    for (var i = 0; i < count; ++i) {
      buf.skip(36);
      var size = buf.getInt();
      var key = null;
      if (size > 0) {
        key = this.transcoder.decode(buf.read(size), 0, size);
        var code = buf.getInt();
        codes[key && key.pkey || key] = code;
      }
    }
  }
  return res;
};


module.exports = PrefixPuts;
