/*!
 * node-tair - lib/packets/put.js
 *
 * Copyright(c) 2012 - 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 *   sunfang1cn <kate.sf@taobao.com>
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:packets:put');
var util = require('util');
var ByteBuffer = require('byte');
var cons = require('../const');
var BasePacket = require('./base');

/**
 * Put packet
 *
 * @param {String|Buffer|Number} key
 * @param {String|Buffer|Number|Boolean} value
 * @param {Transcoder} transcoder
 * @param {Object} options
 *  - {Number} [namespace] defautl is 0
 *  - {Number} [expired] expired time, default is 0, never expired.
 *  - {Number} [version] default is 0
 */
function Put(key, value, transcoder, options) {
  BasePacket.call(this, cons.TAIR_REQ_PUT_PACKET, false);

  this.key = key;
  this.value = value;

  options = options || {};
  this.keyFlag = options.keyFlag || 0;
  this.valueFlag = options.valueFlag || 0;
  this.namespace = options.namespace || 0;
  this.expired = options.expired || 0;
  this.version = options.version || 0;
  this.transcoder = transcoder;

  this.keyBytes = this.transcoder.encode(this.key);
  this.valueBytes = this.transcoder.encode(this.value);
}

util.inherits(Put, BasePacket);

Put.prototype.encode = function (reqId) {
  var keyBytes = this.keyBytes;
  var valueBytes = this.valueBytes;

  // total length: 1 + 2 + 2 + 4 + 36 + key (String length: 4 + bytes.length) + 36 + 4 + value.length
  //                = 45 + 4 + 40 = 89 + key value length
  // format: | 0 (1 byte, body start) | namespace (2 bytes) | version (2 bytes)
  //         | expired (4 bytes)
  //         | key meta flags (36 bytes)
  //         | key length (4 bytes, CString) | key bytes
  //         | value meta flags (36 bytes)
  //         | value length (4 bytes) | value bytes
  var bodyLength = 89 + keyBytes.length + valueBytes.length;

  var totalLength = cons.PACKET_HEADER_SIZE + 4 + bodyLength;
  var bytes = ByteBuffer.allocate(totalLength);
  // request packet format: | header (12 bytes) | body |
  // header: | flag: 6d 42 64 54 | request id (4 bytes) | packet code (4 bytes) |
  bytes.putInt(cons.TAIR_PACKET_FLAG); // packet flag
  bytes.putInt(reqId); // request id
  // | packet code (4 bytes) | body length (4 bytes) | body |
  bytes.putInt(this.code); // 12

  bytes.putInt(bodyLength);

  // body
  bytes.put(0);
  bytes.putShort(this.namespace);
  bytes.putShort(this.version);
  bytes.putInt(this.expired);

  // key
  this.fillMetaFlag(bytes, this.keyFlag);
  bytes.putInt(keyBytes.length);
  bytes.put(keyBytes);

  this.fillMetaFlag(bytes, this.valueFlag);
  bytes.putInt(valueBytes.length);
  bytes.put(valueBytes);
  return bytes.array();
};

module.exports = Put;
