/*!
 * node-tair - lib/packets/transcoder.js
 *
 * Copyright(c) 2012 - 2013 Alibaba Group Holding Limited.
 *
 * Authors:
 *   苏千 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 *
 * Impl Java DefaultTranscoder.js
 */

"use strict";

/**
 * Module dependencies.
 */

var utility = require('utility');
var Long = require('long');
var cons = require('../const');
var MixedKey = require('./mixed_key');
var IncData = require('./inc_data');

var MAX_INT32 = 2147483647;
var MIN_INT32 = -2147483648;

function Transcoder(withHeader) {
  this._compressionThreshold = cons.TAIR_DEFAULT_COMPRESSION_THRESHOLD;
  this._compressionType = cons.TAIR_DEFAULT_COMPRESSION_TYPE;
  this._charset = cons.DEFAULT_CHARSET;
  this.withHeader = withHeader !== false; // TODO: java impl default is true. cpp is false.
  this.COMPRESSIONTYPEOFFSET = 12;
}

Transcoder.prototype.encodeString = function (o) {
  return new Buffer(o, this._charset);
};

Transcoder.prototype.decodeString = function (vb) {
  return vb.toString(this._charset);
};

Transcoder.prototype.encodeInt = function (i) {
  var b = new Buffer(4);
  b.writeInt32BE(i, 0);
  return b;
};

Transcoder.prototype.encodeLong = function (value) {
  var b = new Buffer(8);
  b.writeInt32BE(value.high, 0);
  b.writeInt32BE(value.low, 4);
  return b;
};

Transcoder.prototype.decodeLong = function (vb) {
  var high = vb.readInt32BE(0);
  var low = vb.readInt32BE(4);
  return utility.toSafeNumber(new Long(low, high).toString());
};

Transcoder.prototype.decodeInt = function (vb) {
  return vb.readInt32BE(0);
};

Transcoder.prototype.encode = function (object, prefixCared) {
  // if (object == null) {
  //   throw new IllegalArgumentException("key,value can not be null");
  // }

  if (object instanceof MixedKey && prefixCared) {
    return object.encode(true); //~ only return encoded prefix
  }

  var b = null;
  var flag = 0;
  var toCompress = true;

  var t = typeof object;
  if (t === 'string') {
    b = this.encodeString(object);
    flag = cons.TAIR_STYPE_STRING;
  } else if (object instanceof Long) {
    b = this.encodeLong(object);
    flag = cons.TAIR_STYPE_LONG;
  } else if (t === 'number') {
    if (object > MAX_INT32 || object < MIN_INT32) {
      b = this.encodeLong(Long.fromNumber(object));
      flag = cons.TAIR_STYPE_LONG;
    } else {
      b = this.encodeInt(object);
      flag = cons.TAIR_STYPE_INT;
    }
  } else if (t === 'boolean') {
    b = this.encodeBoolean(object);
    flag = cons.TAIR_STYPE_BOOL;
  // } else if (object instanceof Date) {
  //   b = TranscoderUtil.encodeLong(((Date) object).getTime());
  //   flag = TairConstant.TAIR_STYPE_DATE;
  // } else if (object instanceof Byte) {
  //     b    = TranscoderUtil.encodeByte((Byte) object);
  //     flag = TairConstant.TAIR_STYPE_BYTE;
  // } else if (object instanceof Float) {
  //     b    = TranscoderUtil.encodeInt(Float.floatToRawIntBits((Float) object));
  //     flag = TairConstant.TAIR_STYPE_FLOAT;
  // } else if (object instanceof Double) {
  //     b    = TranscoderUtil.encodeLong(Double.doubleToRawLongBits((Double) object));
  //     flag = TairConstant.TAIR_STYPE_DOUBLE;
  } else if (Buffer.isBuffer(object)) {
    b = object;
    flag = cons.TAIR_STYPE_BYTEARRAY;
  } else if (object instanceof IncData) {
    b = object.encode();
    flag = cons.TAIR_STYPE_INCDATA;
  } else if (object instanceof MixedKey) {
    b = object.encode(false); //~ encode regardless of prefix
    flag = cons.TAIR_STYPE_MIXEDKEY;
    toCompress = false; //~ not to compress mixed key, for the sake of flag consistency
  } else {
  //   b = TranscoderUtil.serialize(object);
  //   flag = TairConstant.TAIR_STYPE_SERIALIZE;
  }

  flag <<= 1;

  if (b.length > this.compressionThreshold && toCompress) {
    b = this.compress(b, this.compressionType);
    var type = this.compressionType;
    type <<= this.COMPRESSIONTYPEOFFSET;
    flag += type;
    flag += 1;
  }

  // TairUtil.checkMalloc(b.length + 2); max size 1MB

  if (false === this.withHeader) {
    return b;
  }

  var fg = new Buffer(2);

  fg[1] = flag & 0xFF;
  fg[0] = (flag >> 8) & 0xFF;
  return Buffer.concat([fg, b], 2 + b.length);
};

Transcoder.prototype.decode = function (data, offset, size, prefixSize) {
  offset = offset || 0;
  size = size || data.length;
  ///////////////////////////////////////////////////////////
  if (false === this.withHeader) {
    // C++ format
    if (data[offset + size - 1] === 0) {
      // remove CString end: '\u0000'
      size--;
    }
    var value = data.slice(offset, size).toString();
    return value;
  }

  prefixSize = prefixSize || 0;
  var index = 0;
  // convert long data
  for (index = 0; index < size && (index + offset < data.length); index++) {
    // +: 43, -: 45
    var v = data[index + offset];
    if (v === 43 || v === 45) {
      continue;
    }
    // '0': 48, '9': 57
    if (v < 48 || v > 57) {
      break;
    }
  }
  if (index === size) {
    var tmpstr = Number(data.toString());
    return tmpstr;
    // try {
    //   long tmp = Long.valueOf(tmpstr);
    //   return tmp;
    // } catch (NumberFormatException e) {

    // }
  }

  var vb = data.slice(2);

  var obj = null;

  var flags = 0;

  for (var i = 0; i < 2; i++) {
    var b = data[offset + i];
    flags = (flags << 8) | ((b < 0) ? (256 + b) : b);
  }

  if ((flags & 1) === 1) {
    // TODO: dont support compress format now, NTBD
    return data.slice(offset, size).toString();

    // the high four bits of flag represent the type of compression
    var compressType = flags >> this.COMPRESSIONTYPEOFFSET;
    vb = this.decompress(vb, compressType);
  }

  var type = (flags >> 1) & 0xF;

  switch (type) {
  case cons.TAIR_STYPE_INT:
    obj = this.decodeInt(vb);
    break;
  case cons.TAIR_STYPE_STRING:
    obj = this.decodeString(vb);
    break;
  // case cons.TAIR_STYPE_BOOL:
  //   obj = this.decodeBoolean(vb);
  //   break;
  case cons.TAIR_STYPE_LONG:
    obj = this.decodeLong(vb);
    break;
  case cons.TAIR_STYPE_DATE:
    obj = new Date(this.decodeLong(vb));
    break;
  // case cons.TAIR_STYPE_BYTE:
  //   obj = TranscoderUtil.decodeByte(vb);
  //   break;
  // case cons.TAIR_STYPE_FLOAT:
  //   Integer f = TranscoderUtil.decodeInt(vb);
  //   obj = new Float(Float.intBitsToFloat(f));
  //   break;
  // case cons.TAIR_STYPE_DOUBLE:
  //   Long l = TranscoderUtil.decodeLong(vb);
  //   obj = new Double(Double.longBitsToDouble(l));
  //   break;
  case cons.TAIR_STYPE_BYTEARRAY:
    obj = vb;
    break;
  // case TairConstant.TAIR_STYPE_SERIALIZE:
  //   obj = TranscoderUtil.deserialize(vb);
  //   break;
  case cons.TAIR_STYPE_INCDATA:
    obj = IncData.decode(vb);
    break;
  case cons.TAIR_STYPE_MIXEDKEY:
    obj = new MixedKey(this).decode(vb, prefixSize);
    break;
  default:
    break;
      // throw new RuntimeException("unknow serialize flag: " + type);
  }

  return obj;
};


module.exports = Transcoder;
