/**!
 * node-tair - lib/servers/base_server.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * Authors:
 *   苏千 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:servers:base_server');
var net = require('net');
var ByteBuffer = require('byte');
var EventEmitter = require('events').EventEmitter;
var util = require('util');
var cons = require('../const');
var Request = require('./request');
var Response = require('./response');
var Transcoder = require('../packets/transcoder');
var utils = require('../utils');

var CPPTranscoder = new Transcoder(false);

/**
 * Base Tair Server
 * @param {Object} options
 *  - {Boolean} [withHeader] true: java format, false: cpp fomat, default is true.
 */
function BaseServer(options) {
  EventEmitter.call(this);

  options = options || {};
  this.transcoder = new Transcoder(options.withHeader !== false);
  this._server = net.createServer();
  this._server.on('connection', this._handleConnection.bind(this));
}

util.inherits(BaseServer, EventEmitter);

BaseServer.prototype._handleConnection = function (socket) {
  var reading = 'header';
  var req;
  var that = this;
  var handleBody = function () {
    var body = socket.read(req.bodySize);
    debug('req#%d try to receive %d bytes, got %d bytes',
      req.id, req.bodySize, body && body.length || 0);
    if (!body) {
      return;
    }

    if (body.length !== req.bodySize) {
      console.warn('[%s] [%s] %s:%s wrong tair body: %d bytes',
        Date(), process.pid, socket.remoteAddress, socket.remotePort, body.length);
      utils.printBuffer(body);
      return socket.destroy();
    }

    var transcoder = that.transcoder;
    req = new Request(req.id, req.code, body, transcoder, socket);
    var res = new Response(req, transcoder, socket);

    debug('emit request: #%s:%s:%j, body %d bytes', req.id, req.name, req.data, body.length);
    if (!req.name) {
      res.sendError({message: 'unsupport packet, code ' + req.code, code: -20008});
    } else if (req.name === 'ping') {
      res.send(req.data);
    } else {
      that.emit('request', req, res);
    }

    req = null;
    reading = 'header';
    handleHeader();
  };

  var handleHeader = function () {
    // flag, req id, packet code, body length = 16 bytes
    var header = socket.read(16);
    if (!header) {
      return;
    }

    if (header.length !== 16) {
      console.warn('[%s] [%s] %s:%s wrong tair header: %d bytes',
        Date(), process.pid, socket.remoteAddress, socket.remotePort, header.length);
      utils.printBuffer(header);
      return socket.destroy();
    }

    var buf = ByteBuffer.wrap(header);

    var flag = buf.getInt();
    if (flag !== cons.TAIR_PACKET_FLAG) {
      // TODO: log this
      return socket.destroy();
    }

    var reqId = buf.getInt();
    var packetCode = buf.getInt();
    var length = buf.getInt();

    debug('req#%d got response header, length: %d, packetCode: %d, body length: %d',
      reqId, header.length, packetCode, length);

    req = {
      id: reqId,
      code: packetCode,
      bodySize: length
    };

    reading = 'body';
    handleBody();
  };

  var handleReponse = function () {
    if (reading === 'header') {
      handleHeader();
    } else {
      handleBody();
    }
  };
  socket.on('readable', handleReponse);
  socket.on('error', function (err) {
    console.error('[%s] [%s] tair server: client socket error: %s', Date(), process.pid, err);
  });
};

BaseServer.prototype.listen = function (port, callback) {
  this._server.listen(port, callback);
};

BaseServer.prototype.close = function (callback) {
  this._server.close(callback);
};


module.exports = BaseServer;
