/**!
 * node-tair - lib/servers/data_node.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * Authors:
 *   苏千 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var debug = require('debug')('tair:servers:data_node');
var util = require('util');
var address = require('address');
var BaseServer = require('./base_server');
var Connection = require('../connection');
var DataNodeHeartBeat = require('../packets').DataNodeHeartBeat;

/**
 * DataNode Server
 * @param {Object} options
 *  - {Array} configServers e.g.: `[{host: '127.0.0.1', port: 1234}, {host: '127.0.0.1', port: 4567}]`
 *  - {String} group
 *  - {Number} [heartbeatInterval] default is 10000 ms
 *  - {Boolean} [withHeader] true: java format, false: cpp fomat, default is true.
 */
function DataNode(options) {
  BaseServer.call(this, options);
  this._configServers = options.configServers;
  this._group = options.group;
  this._host = options.host || address.ip();
  this._port = 0;
  this._heartbeatInterval = options.heartbeatInterval || 10000;

  // key: method + namesapce : handles
  this._handles = {};
  this._handleCount = 0;
  this._beforeMiddlewares = [];
  this._afterMiddlewares = [];
  this.on('request', this._handleRequest.bind(this));
}

util.inherits(DataNode, BaseServer);

DataNode.prototype.listen = function (port, callback) {
  var that = this;
  that._port = port;
  BaseServer.prototype.listen.call(that, port, function (err) {
    if (err) {
      return callback && callback(err);
    }
    that._init();
    callback && callback();
  });
};

/**
 * Routing methods
 */

var methods = ['get'];

/**
 * Handle `method` request
 * @param {Number} namespace
 * @param {Function(req, res, next)} handle
 * @return {DataNode} this
 */
methods.forEach(function (method) {
  DataNode.prototype[method] = function (namespace, handle) {
    this._handleCount++;
    var name = method + ':' + namespace;
    var handles = this._handles[name];
    if (!handles) {
      handles = this._handles[name] = [];
    }
    handles.push(handle);
    return this;
  };
});

/**
 * Add middleware just like connect middleware.
 *
 * @param {Function([err, ]req, res, next)} middleware
 * @return {DataNode} this
 */
DataNode.prototype.use = function (middleware) {
  if (this._handleCount === 0) {
    this._beforeMiddlewares.push(middleware);
  } else {
    this._afterMiddlewares.push(middleware);
  }
  return this;
};

DataNode.prototype._handleRequest = function handleRequest(req, res) {
  var namespace = req.data && req.data.namespace || '-1';
  var name = req.name + ':' + namespace;
  var handles = this._handles[name] || [];
  if (this._beforeMiddlewares.length > 0) {
    handles = this._beforeMiddlewares.concat(handles);
  }
  if (this._afterMiddlewares.length > 0) {
    handles = handles.concat(this._afterMiddlewares);
  }
  var handleIndex = 0;
  var next = function (err) {
    var handle = handles[handleIndex++];
    if (handle) {
      var argsLength = handle.length;
      if (err) {
        if (argsLength === 4) {
          handle(err, req, res, next);
        } else {
          next(err);
        }
      } else {
        if (argsLength < 4) {
          handle(req, res, next);
        } else {
          next();
        }
      }
    } else {
      // no handler
      if (err) {
        err.code = err.code || -3999;
        res.sendError(err);
        return;
      }
      // not support
      res.sendError({message: '404 not support', code: -20008, status: 404});
    }
  };

  next();
};

DataNode.prototype._init = function () {
  debug('data init, %s:%s:%s', this._group, this._host, this._port);
  var that = this;
  that._configServers.forEach(function (cs) {
    var conn = new Connection({
      host: cs.host,
      port: cs.port,
    });

    var heartbeat = function () {
      var connName = conn && conn.name;
      conn && conn.call(new DataNodeHeartBeat(that._group, that._host, that._port), function (err, res) {
        if (err || !res || !res.body || res.body.code < 0) {
          console.warn('[%s] [%s] connection:%s data node (%s, %s, %s) heartbeat error, %s, %j',
            Date(), process.pid, connName, that.group, that.host, that.port, err, res);
          // close the connection.
          conn && conn.close();
        }
        debug('send %s:%s:%s heartbeat success', that._group, that._host, that._port);
      });
    };

    var handleClose = function () {
      conn = null;
      // reconnect
      setTimeout(function () {
        conn = new Connection({
          host: cs.host,
          port: cs.port,
        });
        conn.once('connect', heartbeat);
        conn.once('close', handleClose);
      }, 1000);
    };
    conn.once('close', handleClose);

    var heartbeatTimer = setInterval(heartbeat, that._heartbeatInterval);
    heartbeat();
  });
};

module.exports = DataNode;
