/*!
 * node-tair - test/client.test.js
 *
 * Copyright(c) 2012 - 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 *   sunfang1cn <kate.sf@taobao.com>
 */

"use strict";

/**
 * Module dependencies.
 */

var pedding = require('pedding');
var fs = require('fs');
var path = require('path');
var should = require('should');
var mm = require('mm');
var Connection = require('../lib/connection');
var Tair = require('../');
var fixtures = path.join(__dirname, 'fixtures');

describe('client.test.js', function () {
  var tair;
  var mdbTair;

  var fooValue = 'node tair foo result: bar, Date: ' + Date();

  before(function (done) {
    tair = new Tair({
      dataId: 'ldbcommon-daily',
      diamond: 'commonconfig.taobao.net:8080',
      namespace: 171,
    }).on('connect', function () {
      var primaryDone = false;
      var backupDone = false;
      mdbTair = new Tair({
        // dataId: 'market2-daily',
        dataId: 'tmall2-daily',
        diamond: 'commonconfig.taobao.net:8080',
        namespace: 171,
      }).on('connect', done);
    });
  });

  afterEach(mm.restore);

  after(function (done) {
    done = pedding(2, done);
    tair.close();
    mdbTair.close();
    tair.on('close', done);
    mdbTair.on('close', done);
  });

  describe('Tair()', function () {
    it('should load downServers from config', function (done) {
      var tair = Tair({
        servers: [{host: '10.232.12.57', port: 5198}],
        group: 'group_1',
      });
      tair.on('connect', function () {
        tair.get('1', function (err) {
          should.not.exist(err);
          tair.get('foo', function (err) {
            should.exist(err);
            err.name.should.equal('TairDataNodeNotAliveError');
            done();
          });
        });
      });
    });

    it('should work with options.servers', function (done) {
      var tair = new Tair({
        group: 'group_market2',
        servers: [{host: '10.235.144.195', port: 5198}],
        namespace: 171
      });
      tair.on('connect', function () {
        tair.close();
        done();
      });
    });

    it('should return TairDataNodeNotAliveError when tair client not ready', function (done) {
      var tair = new Tair({
        group: 'group_market2',
        servers: [{host: '10.235.144.195', port: 5198}]
      });
      tair.get('foo', function (err, data) {
        should.exist(err);
        err.name.should.equal('TairDataNodeNotAliveError');
        done();
      });
    });

    it('should return TairDataNodeNotAliveError when tair client not ready', function (done) {
      var tair = Tair({
        dataId: 'market2-daily',
        diamond: 'commonconfig.taobao.net:8080',
        namespace: 171,
      }).on('connect', function () {
        mm(tair._clusterMap[tair._clusterNames[0]].client._groupInfo, 'nodes', {});
        tair.get('foo', function (err, data) {
          should.exist(err);
          err.name.should.equal('TairDataNodeNotAliveError');
          done();
        });
      });
    });
  });

  describe('lock(), unlock()', function () {
    var key = 'nodejs-lock-' + Date();
    it('should lock a key', function (done) {
      mdbTair.lock(key, 2, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        // lock again should fail
        mdbTair.lock(key, 2, function (err, success, res) {
          should.exist(err);
          err.name.should.equal('TairRequestError');
          err.message.should.equal('-3975: lock already exist');
          success.should.equal(false);
          // should unlock success
          mdbTair.unlock(key, function (err, success, res) {
            should.not.exist(err);
            success.should.equal(true);
            // should unlock success whatever unlock
            mdbTair.unlock(key, function (err, success, res) {
              should.not.exist(err);
              success.should.equal(true);
              done();
            });
          });
        });
      });
    });

    it('should lock a key 1 seconds, after that, lock success', function (done) {
      var key2 = key + '2';
      mdbTair.lock(key2, 1, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        // lock again should fail
        mdbTair.lock(key2, 1, function (err, success, res) {
          should.exist(err);
          err.name.should.equal('TairRequestError');
          err.message.should.equal('-3975: lock already exist');
          success.should.equal(false);
          setTimeout(function () {
            mdbTair.lock(key2, 2, function (err, success, res) {
              should.not.exist(err);
              success.should.equal(true);
              mdbTair.unlock(key2, function (err, success) {
                should.not.exist(err);
                success.should.equal(true);
                done();
              });
            });
          }, 2000);
        });
      });
    });

    it('should unlock a expired key success', function (done) {
      var key3 = key + '3';
      mdbTair.lock(key3, 1, function (err, success) {
        should.not.exist(err);
        success.should.equal(true);
        setTimeout(function () {
          mdbTair.unlock(key3, function (err, success, res) {
            should.not.exist(err);
            success.should.equal(true);
            res.body.code.should.equal(-3988); // data expired
            done();
          });
        }, 2000);
      });
    });
  });

  describe('incr(), decr()', function () {
    var startValue = 0;
    beforeEach(function (done) {
      mdbTair.get('node-tair-incr1', function (err, value, res) {
        should.not.exist(err);
        startValue = value || 0;
        console.log('startValue: %s', value);
        done();
      });
    });

    it('should incr 1 and incr 10 forever', function (done) {
      mdbTair.incr('node-tair-incr1', 1, function (err, value, res) {
        should.not.exist(err);
        // 隔一段时候后, 再incr这个key, 返回的value会不等于真正的值.
        value.should.equal(startValue + 1);
        startValue = value;
        res.body.should.have.keys('configVersion', 'value');
        mdbTair.incr('node-tair-incr1', 101, function (err, value, res) {
          should.not.exist(err);
          value.should.equal(startValue + 101);
          res.body.should.have.keys('configVersion', 'value');
          startValue = value;
          done();
        });
      });
    });

    it('should decr 1', function (done) {
      mdbTair.decr('node-tair-incr1', 1, function (err, value, res) {
        should.not.exist(err);
        value.should.equal(startValue - 1);
        startValue = value;
        res.body.should.have.keys('configVersion', 'value');
        mdbTair.decr('node-tair-incr1', 100, function (err, value, res) {
          should.not.exist(err);
          value.should.equal(startValue - 100);
          res.body.should.have.keys('configVersion', 'value');
          startValue = value;
          done();
        });
      });
    });

    it('should incr 4 with default 5 and expired after 1 seconds', function (done) {
      mdbTair.incr('node-tair-incr-default-5', 4, {defaultValue: 5, expired: 1},
      function (err, value, res) {
        should.not.exist(err);
        value.should.equal(9);
        startValue = value;
        res.body.should.have.keys('configVersion', 'value');
        mdbTair.get('node-tair-incr-default-5', function (err, value, res) {
          should.not.exist(err);
          value.should.equal(9);
          setTimeout(function () {
            mdbTair.get('node-tair-incr-default-5', function (err, value) {
              should.not.exist(err);
              should.not.exist(value);
              done();
            });
          }, 2000);
        });
      });
    });
  });

  describe('setCount()', function () {
    it('should set a counter', function (done) {
      mdbTair.setCount('nodejs-tair-counter', 999, function (err, success) {
        should.not.exist(err);
        success.should.equal(true);
        mdbTair.get('nodejs-tair-counter', function (err, value) {
          should.not.exist(err);
          value.should.equal(999);
          done();
        });
      });
    });
  });

  describe('put()', function () {
    it('should put key:foo value:bar success', function (done) {
      // put packet: old: <Buffer 6d 42 64 54 0 0 0 5 0 0 0 1 0 0 0 74 0| 0 0| 0 0| 0 0 0 0| 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 e 6e 6f 64 65 2d 74 61 69 72 2d 66 6f 6f 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 d 6e 6f 64 65 2d 74 61 69 72 2d 62 61 72>
      //             new: <Buffer 6d 42 64 54 0 0 0 5 0 0 0 1 0 0 0 6f 0| 0 0| 0 0| 0 0 0 1| 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 e 6e 6f 64 65 2d 74 61 69 72 2d 66 6f 6f 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 d 6e 6f 64 65 2d 74 61 69 72 2d 62 61 72>
      // reponse packet: <Buffer 6d 42 64 54 0 0 0 5 0 0 0 65 0 0 0 c 0 0 0 7d 0 0 0 0 0 0 0 0>
      tair.put('node-tair-foo', fooValue, function (err, success, res) {
        should.not.exist(err);
        should.exist(res);
        success.should.equal(true);
        res.body.should.have.property('code', 0);
        done();
      });
    });

    it('should mock put error', function (done) {
      mm.error(tair._clusterMap[tair._clusterNames[0]].client, '_send');
      tair.put('node-tair-foo', fooValue, function (err, success, res) {
        should.exist(err);
        err.message.should.equal('mm mock error');
        should.not.exist(success);
        done();
      });
    });

    it('should put success and read from backup cluster', function (done) {
      var backupValue = fooValue + ', now: ' + new Date();
      mdbTair.put('node-tair-foo-read-from-backup', backupValue, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        var c1 = mdbTair._clusterNames[0];
        var c2 = mdbTair._clusterNames[1];
        mdbTair._clusterMap[c1].client._groupInfo.nodes.should.not.eql(mdbTair._clusterMap[c2].client._groupInfo.nodes);
        mm(mdbTair._clusterMap[c1].client._groupInfo, 'nodes', {});
        mdbTair.get('node-tair-foo-read-from-backup', function (err, value) {
          should.not.exist(err);
          should.exist(value);
          value.should.equal(backupValue);
          done();
        });
      });
    });

    it('should mock put response error', function (done) {
      mm.error(Connection.prototype, 'call');
      tair.put('node-tair-foo', fooValue, function (err, success, res) {
        should.exist(err);
        err.message.should.equal('mm mock error');
        should.not.exist(success);
        done();
      });
    });

    it('should mock put return response code error', function (done) {
      var Connection = require('../lib/connection');
      mm.data(Connection.prototype, 'call', {body: {code: -1}});
      tair.put('node-tair-foo', fooValue, function (err, success, res) {
        should.exist(err);
        err.name.should.equal('TairRequestError');
        err.message.should.equal('-1: connection error or timeout');
        should.not.exist(success);
        done();
      });
    });

    it('should put key:Buffer(foo) value:bar success', function (done) {
      tair.put(new Buffer('node-tair-foo-buffer'), fooValue, function (err, success, res) {
        should.not.exist(err);
        should.exist(res);
        success.should.equal(true);
        res.body.should.have.property('code', 0);
        tair.get(new Buffer('node-tair-foo-buffer'), function (err, data) {
          should.not.exist(err);
          should.exist(data);
          data.toString().should.equal(fooValue);
          done();
        });
      });
    });

    it('should put big data success', function (done) {
      var content = fs.readFileSync(__filename, 'utf8');
      tair.put('node-tair-big-data', content, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        res.body.code.should.equal(0);
        tair.get('node-tair-big-data', function (err, data) {
          should.not.exist(err);
          data.toString().should.equal(content);
          done();
        });
      });
    });

    it('should set a timestamp', function (done) {
      var v = new Date().getTime();
      tair.put('node-test-timestamp', v, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        tair.get('node-test-timestamp', function (err, value) {
          should.not.exist(err);
          value.should.equal(v);
          done();
        });
      });
    });

    it('should put a value without callback', function (done) {
      tair.put('node-without-callback', 123);
      setTimeout(done, 500);
    });

    it('should set int', function (done) {
      done = pedding(3, done);
      tair.put('node-test-2147483647', 2147483647, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        tair.get('node-test-2147483647', function (err, v) {
          should.not.exist(err);
          v.should.equal(2147483647);
          done();
        });
      });

      tair.put('node-test--2147483648', -2147483648 , function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        tair.get('node-test--2147483648', function (err, v, res) {
          should.not.exist(err);
          v.should.equal(-2147483648);
          done();
        });
      });

      tair.put('node-test-2147483648', 2147483648 , function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        tair.get('node-test-2147483648', function (err, v) {
          should.not.exist(err);
          v.should.equal(2147483648);
          done();
        });
      });
    });

    it('should ldb batch put and get value success', function (done) {
      var count = 1000;
      var initCount = 20;
      done = pedding(count + initCount, done);

      var putOne = function (i) {
        var key = 'node-tair-batch-put-' + i;
        tair.put('node-tair-batch-put-' + i, fooValue + '-value: ' + i, function (err, success, res) {
          should.not.exist(err);
          should.exist(res);
          success.should.equal(true);
          res.body.should.have.property('code', 0);
          tair.get(key, function (err, value, res) {
            should.not.exist(err);
            should.exist(value);
            // console.log(i, value.toString(), 'use ' + res.use + 'ms')
            value.toString().should.equal(fooValue + '-value: ' + i);
            done();
          });
        });
      };
      for (var i = 0; i < initCount; i++) {
        putOne(i);
      }
      setTimeout(function () {
        for (var i = 0; i < count; i++) {
          putOne(i);
        }
      }, 100);
    });

    it('should mdb batch put and get value success', function (done) {
      var count = 1000;
      var initCount = 20;
      done = pedding(count + initCount, done);

      var putOne = function (i) {
        var key = 'node-tair-batch-put-' + i;
        mdbTair.put('node-tair-batch-put-' + i, fooValue + '-value: ' + i, function (err, success, res) {
          should.not.exist(err);
          should.exist(res);
          success.should.equal(true);
          res.body.should.have.property('code', 0);
          mdbTair.get(key, function (err, value, res) {
            should.not.exist(err);
            should.exist(value);
            // console.log(i, value.toString(), 'use ' + res.use + 'ms')
            value.toString().should.equal(fooValue + '-value: ' + i);
            done();
          });
        });
      };
      for (var i = 0; i < initCount; i++) {
        putOne(i);
      }
      setTimeout(function () {
        for (var i = 0; i < count; i++) {
          putOne(i);
        }
      }, 100);
    });

    it('should put expired data', function (done) {
      tair.put('node-tair-expired-3', 'expired after 3 seconds', {expired: 3}, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        res.body.code.should.equal(0);
      });
      tair.put('node-tair-expired-1', 'expired after 1 second', {expired: 1}, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        res.body.code.should.equal(0);
        setTimeout(function () {
          done = pedding(2, done);
          tair.get('node-tair-expired-1', function (err, data) {
            should.not.exist(err);
            data = data && data.toString();
            should.not.exist(data);
            done();
          });
          tair.get('node-tair-expired-3', function (err, data) {
            should.not.exist(err);
            data = data && data.toString();
            should.exist(data);
            data.should.equal('expired after 3 seconds');
            done();
          });
        }, 2000);
      });
    });
  });

  describe('ldb invalid()', function () {
    var key = 'node-tair-invalid-test';
    var value = key + Date();
    beforeEach(function (done) {
      tair.put(key, value, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        tair.get(key, function (err, data, res) {
          should.not.exist(err);
          data.should.be.a.string;
          data.should.equal(value);
          done();
        });
      });
    });

    it('should remove not exists data success', function (done) {
      done = pedding(11, done);
      tair.invalid(key + 'not', function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        res.body.should.have.property('code');
        tair.get(key + 'not', function (err, data) {
          should.not.exist(err);
          should.not.exist(data);
          done();
        });
      });

      for (var i = 0; i < 10; i++) {
        tair.invalid(key + 'not', function (err, success) {
          should.not.exist(err);
          success.should.equal(true);
          done();
        });
      }
    });

    it('should remove exists data success', function (done) {
      tair.invalid(key, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        should.exist(res);
        res.body.code.should.equal(0);
        res.body.count.should.equal(0);
        tair.get(key, function (err, data, res) {
          should.not.exist(err);
          should.not.exist(data);
          done();
        });
      });
    });

    it('should return mock error', function (done) {
      mm.error(tair._clusterMap[tair._clusterNames[0]].client, '_send');
      tair.invalid(key, function (err, success) {
        should.exist(err);
        should.ok(!success);
        done();
      });
    });
  });

  describe('mdb invalid()', function () {
    var key = 'node-tair-invalid-test';
    var value = key + Date();
    beforeEach(function (done) {
      mdbTair.put(key, value, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        mdbTair.get(key, function (err, data, res) {
          should.not.exist(err);
          data.should.be.a.string;
          data.should.equal(value);
          done();
        });
      });
    });

    it('should remove not exists data success', function (done) {
      done = pedding(11, done);
      mdbTair.invalid(key + 'not', function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        res.body.should.have.property('code');
        mdbTair.get(key + 'not', function (err, data) {
          should.not.exist(err);
          should.not.exist(data);
          done();
        });
      });

      for (var i = 0; i < 10; i++) {
        mdbTair.invalid(key + 'not', function (err, success) {
          should.not.exist(err);
          success.should.equal(true);
          done();
        });
      }
    });

    it('should remove exists data success', function (done) {
      mdbTair.invalid(key, function (err, success, res) {
        should.not.exist(err);
        success.should.equal(true);
        should.exist(res);
        res.body.code.should.equal(0);
        mdbTair.get(key, function (err, data, res) {
          should.not.exist(err);
          should.not.exist(data);
          done();
        });
      });
    });

    it('should return mock error', function (done) {
      for (var k in mdbTair._clusterMap) {
        mm.error(mdbTair._clusterMap[k].client, '_send');
      }
      mdbTair.invalid(key, function (err, success) {
        should.exist(err);
        should.ok(!success);
        done();
      });
    });
  });

  describe('prefixPuts()', function () {
    it('should prefix put keyValues', function (done) {
      var keyValues = [
        ['kk1', '1'],
        ['kk2', 2],
        ['kk3', 'k3']
      ];
      tair.prefixPuts('nodejs-prefixPuts-test', keyValues, function (err, data, res) {
        should.not.exist(err);
        data.should.equal(true);
        should.exist(data);
        tair.prefixGets('nodejs-prefixPuts-test', ['kk1', 'kk2', 'kk3'], function (err, data) {
          should.not.exist(err);
          data.should.eql({
            kk1: '1',
            kk2: 2,
            kk3: 'k3'
          });
          done();
        });
      });
    });

    it('should mock error', function (done) {
      for (var k in tair._clusterMap) {
        mm.error(tair._clusterMap[k].client, '_send');
      }
      var keyValues = [
        ['kk1', '1'],
        ['kk2', 2],
        ['kk3', 'k3']
      ];
      tair.prefixPuts('nodejs-prefixPuts-test', keyValues, function (err, data, res) {
        should.exist(err);
        err.message.should.equal('mm mock error');
        done();
      });
    });
  });

  describe('prefixGet()', function () {
    before(function (done) {
      done = pedding(2, done);
      mdbTair.prefixPut('320', 'uidstr', '320string', done);
      mdbTair.prefixPut('320', 'uid', 320, done);
    });

    it('should get 320:uidstr return string', function (done) {
      mdbTair.prefixGet('320', 'uidstr', function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.a.string;
        data.should.equal('320string');
        res.body.should.have.keys('version', 'code', 'count', 'key', 'value', 'meta');
        done();
      });
    });

    it('should get 320:uid return Number', function (done) {
      mdbTair.prefixGet('320', 'uid', function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.a.string;
        data.should.equal(320);
        res.body.should.have.keys('version', 'code', 'count', 'key', 'value', 'meta');
        done();
      });
    });

    it('should mock primary clusters data nodes all not alive, and backup is empty', function (done) {
      for (var k in tair._clusterMap) {
        mm(tair._clusterMap[k].client._groupInfo, 'servers', [[]]);
      }
      tair.prefixGet('320', 'uidstr', function (err, data, res) {
        should.exist(err);
        err.name.should.equal('TairDataNodeNotAliveError');
        should.not.exist(data);
        done();
      });
    });

    it('should mock primary and backup clusters data nodes all not alive', function (done) {
      for (var k in mdbTair._clusterMap) {
        mm(mdbTair._clusterMap[k].client._groupInfo, 'servers', [[]]);
      }
      mdbTair.prefixGet('320', 'uidstr', function (err, data, res) {
        should.exist(err);
        err.name.should.equal('TairDataNodeNotAliveError');
        should.not.exist(data);
        done();
      });
    });

    it('should mock primary cluster data nodes all not alive and get data from backup cluster', function (done) {
      for (var k in mdbTair._clusterMap) {
        mm(mdbTair._clusterMap[k].client._groupInfo, 'servers', [[]]);
        break;
      }
      mdbTair.prefixGet('320', 'uidstr', function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.a.string;
        data.should.equal('320string');
        res.body.should.have.keys('version', 'code', 'count', 'key', 'value', 'meta');
        done();
      });
    });

    it('should mock primary cluster data nodes all down and get data from backup cluster', function (done) {
      var downServers = {};
      var client = mdbTair._clusterMap[mdbTair._clusterNames[0]].client;
      var nodes = client._groupInfo.nodes;
      for (var k in nodes) {
        var node = nodes[k];
        downServers[node.host + ':' + node.port] = true;
      }

      mm(client._groupInfo, 'downServers', downServers);
      mdbTair.prefixGet('320', 'uidstr', function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.a.string;
        data.should.equal('320string');
        res.body.should.have.keys('version', 'code', 'count', 'key', 'value', 'meta');
        done();
      });
    });

    it('should mock primary and backup clusters data nodes all down', function (done) {
      var downServers = {};
      var client = mdbTair._clusterMap[mdbTair._clusterNames[0]].client;
      var nodes = client._groupInfo.nodes;
      for (var k in nodes) {
        var node = nodes[k];
        downServers[node.host + ':' + node.port] = true;
      }

      mm(client._groupInfo, 'downServers', downServers);

      downServers = {};
      var client = mdbTair._clusterMap[mdbTair._clusterNames[1]].client;
      var nodes = client._groupInfo.nodes;
      for (var k in nodes) {
        var node = nodes[k];
        downServers[node.host + ':' + node.port] = true;
      }

      mm(client._groupInfo, 'downServers', downServers);
      mdbTair.prefixGet('320', 'uidstr', function (err, data, res) {
        should.exist(err);
        err.name.should.equal('TairDataNodeNotAliveError');
        done();
      });
    });

    it('should get data from copy 2 datanode when copy 1 empty', function (done) {
      var client = mdbTair._clusterMap[mdbTair._clusterNames[0]].client;
      var servers = client._groupInfo.servers;
      mm(client._groupInfo, 'servers', [[], servers[0]]);
      mm(client._groupInfo, 'copy', 2);
      mdbTair.prefixGet('320', 'uidstr', function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.a.string;
        data.should.equal('320string');
        res.body.should.have.keys('version', 'code', 'count', 'key', 'value', 'meta');
        done();
      });
    });

    it('should get data from backup copy 2 datanode when others are all downs', function (done) {
      var client = mdbTair._clusterMap[mdbTair._clusterNames[0]].client;
      var servers = client._groupInfo.servers;
      var downServers = {};
      for (var k in client._groupInfo.nodes) {
        var node = client._groupInfo.nodes[k];
        downServers[node.host + ':' + node.port] = true;
      }
      mm(client._groupInfo, 'downServers', downServers);
      mm(client._groupInfo, 'servers', [client._groupInfo.servers[0], client._groupInfo.servers[0]]);
      mm(client._groupInfo, 'copy', 2);

      client = mdbTair._clusterMap[mdbTair._clusterNames[1]].client;
      mm(client._groupInfo, 'servers', [[], client._groupInfo.servers[0]]);
      mm(client._groupInfo, 'copy', 2);
      mdbTair.prefixGet('320', 'uidstr', function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.a.string;
        data.should.equal('320string');
        res.body.should.have.keys('version', 'code', 'count', 'key', 'value', 'meta');
        done();
      });
    });
  });

  describe('prefixGets()', function () {
    it('should get 320:uidstr,uid return obj', function (done) {
      mdbTair.prefixGets('320', ['uidstr', 'uid'], function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.an.object;
        data.uid.should.equal(320);
        data.uidstr.should.equal('320string');
        data.should.eql({
          uid: 320,
          uidstr: '320string'
        });
        res.body.codes.should.eql({ uid: 0, uidstr: 0 });
        res.body.successCount.should.equal(2);
        res.body.failedCount.should.equal(0);
        res.body.should.have.keys('version', 'code', 'successCount', 'failedCount',
          'key', 'value', 'codes', 'originalCode');
        done();
      });
    });

    it('should get 320:uidstr,uid,not-exists return obj', function (done) {
      mdbTair.prefixGets('320', ['uidstr', 'uid', 'not-exists'], function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.an.object;
        data.uid.should.equal(320);
        data.uidstr.should.equal('320string');
        data.should.eql({
          uid: 320,
          uidstr: '320string'
        });
        res.body.codes.should.eql({ uid: 0, uidstr: 0, 'not-exists': -3998 });
        res.body.successCount.should.equal(2);
        res.body.failedCount.should.equal(1);
        res.body.should.have.keys('version', 'code', 'successCount', 'failedCount',
          'key', 'value', 'codes', 'originalCode');
        done();
      });
    });

    it('should get 320:uidstr return obj', function (done) {
      mdbTair.prefixGets('320', ['uidstr'], function (err, data, res) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.an.object;
        data.should.eql({
          uidstr: '320string'
        });
        res.body.codes.should.eql({ uidstr: 0 });
        res.body.successCount.should.equal(1);
        res.body.failedCount.should.equal(0);
        res.body.should.have.keys('version', 'code', 'successCount', 'failedCount',
          'key', 'value', 'codes', 'originalCode');
        done();
      });
    });

    it('should get 320:not-exists,nouid return null', function (done) {
      mdbTair.prefixGets('320', ['not-exists', 'nouid'], function (err, data, res) {
        should.not.exist(err);
        should.not.exist(data);
        res.body.codes.should.eql({ nouid: -3998, 'not-exists': -3998 });
        res.body.successCount.should.equal(0);
        res.body.failedCount.should.equal(2);
        res.body.should.have.keys('version', 'code', 'successCount', 'failedCount',
          'key', 'value', 'codes', 'originalCode');
        done();
      });
    });

    it('should mock error', function (done) {
      for (var k in mdbTair._clusterMap) {
        mm.error(mdbTair._clusterMap[k].client, '_send');
      }
      mdbTair.prefixGets('320', ['uidstr'], function (err, data, res) {
        should.exist(err);
        err.message.should.equal('mm mock error');
        should.not.exist(data);
        done();
      });
    });
  });

  describe('getRangeOnlyValue(), getRange(), getRangeOnlyKey()', function () {
    it('should top 100 rows key values', function (done) {
      tair.getRange('320320320', '8616882133', '8616882233', 0, 10, function (err, items, res) {
        should.not.exist(err);
        should.exist(items);
        items.should.length(res.body.count);
        items.should.eql([
          { key: '8616882133', value: '0|0|10000' },
          { key: '8616882134', value: '1|1|9999' },
          { key: '8616882135', value: '2|2|9998' },
          { key: '8616882136', value: '3|3|9997' },
          { key: '8616882137', value: '0|4|9996' },
          { key: '8616882138', value: '1|5|9995' },
          { key: '8616882139', value: '2|6|9994' },
          { key: '8616882140', value: '3|7|9993' },
          { key: '8616882141', value: '0|8|9992' },
          { key: '8616882142', value: '1|9|9991' }
        ]);
        res.packetCode.should.equal(19); // get range res code
        res.body.should.have.keys('version', 'code', 'items', 'count');
        done();
      });
    });

    it('should top 100 rows values', function (done) {
      tair.getRangeOnlyValue('320320320', '8616882133', '8616882233', 0, 10, function (err, items, res) {
        should.not.exist(err);
        should.exist(items);
        items.should.length(res.body.count);
        items.should.eql([
          { key: null, value: '0|0|10000' },
          { key: null, value: '1|1|9999' },
          { key: null, value: '2|2|9998' },
          { key: null, value: '3|3|9997' },
          { key: null, value: '0|4|9996' },
          { key: null, value: '1|5|9995' },
          { key: null, value: '2|6|9994' },
          { key: null, value: '3|7|9993' },
          { key: null, value: '0|8|9992' },
          { key: null, value: '1|9|9991' }
        ]);
        res.packetCode.should.equal(19); // get range res code
        res.body.should.have.keys('version', 'code', 'items', 'count');
        done();
      });
    });

    it('should top 100 rows keys', function (done) {
      tair.getRangeOnlyKey('320320320', '8616882133', '8616882233', 0, 10, function (err, items, res) {
        should.not.exist(err);
        should.exist(items);
        items.should.length(res.body.count);
        items.should.eql([
          { key: '8616882133', value: null },
          { key: '8616882134', value: null },
          { key: '8616882135', value: null },
          { key: '8616882136', value: null },
          { key: '8616882137', value: null },
          { key: '8616882138', value: null },
          { key: '8616882139', value: null },
          { key: '8616882140', value: null },
          { key: '8616882141', value: null },
          { key: '8616882142', value: null }
        ]);
        res.packetCode.should.equal(19); // get range res code
        res.body.should.have.keys('version', 'code', 'items', 'count');
        done();
      });
    });
  });

  describe('get()', function () {
    it('should get foo success', function (done) {
      tair.get('node-tair-foo', function (err, data) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.a.string;
        data.should.equal(fooValue);
        done();
      });
    });

    it('should get foo java data success', function (done) {
      tair.get('node-tair-foo-java', function (err, data) {
        should.not.exist(err);
        should.exist(data);
        data.should.be.a.string;
        data.should.equal('java-value');
        done();
      });
    });

    it('should get node-tair-benchmark-key Chinese value success', function (done) {
      tair.put('node-tair-benchmark-key', '哈哈', function (err, success) {
        should.not.exist(err);
        success.should.equal(true);
        tair.get('node-tair-benchmark-key', function (err, data) {
          should.not.exist(err);
          should.exist(data);
          data.should.be.a.string;
          data.should.equal('哈哈');
          done();
        });
      });
    });

    it('should return err when _send() error', function (done) {
      for (var k in tair._clusterMap) {
        mm.error(tair._clusterMap[k].client, '_send');
      }
      tair.get('error', function (err, data) {
        should.exist(err);
        done();
      });
    });

  });

  describe('_getGroup()', function () {
    it('should mock get group info from slave', function (done) {
      mm(Connection.prototype, 'call', function (packet, callback) {
        mm.restore();
        process.nextTick(function () {
          callback(new Error('mock error'));
        });
      });
      var client = tair._clusterMap[tair._clusterNames[0]].client;
      client._getGroup(client._version, client._groupName, client._clusterName, function (err) {
        should.not.exist(err);
        client._getGroup(client._version, client._groupName, client._clusterName, done);
      });
    });

    it('should mock get group info error', function (done) {
      var client = tair._clusterMap[tair._clusterNames[0]].client;
      mm.error(Connection.prototype, 'call');
      client._getGroup(client._version, client._groupName, client._clusterName, function (err) {
        should.exist(err);
        err.message.should.equal('mm mock error');
        done();
      });
    });
  });

  describe('connectionHeartbeat()', function () {
    before(function (done) {
      tair.put('connectionHeartbeat-test-key', 'connectionHeartbeat', done);
    });

    it('should mock heartbeat fail and close the connection', function (done) {
      var count = 0;
      var client = tair._clusterMap[tair._clusterNames[0]].client;
      for (var k in client._connections) {
        count++;
        var conn = client._connections[k];
        mm.error(conn, 'call');
      }
      count.should.above(0);
      var doneOne = pedding(count, function (err) {
        should.not.exist(err);
        tair.get('connectionHeartbeat-test-key', function (err, value) {
          should.not.exist(err);
          value.should.equal('connectionHeartbeat');
          done();
        });
      });
      for (var k in client._connections) {
        count++;
        var conn = client._connections[k];
        conn.on('close', function () {
          doneOne();
        });
        conn._heartbeat();
      }
    });
  });

  describe('_handleClusterConfig()', function () {
    it('should handle json parse error', function () {
      var tair = new Tair({
        dataId: 'ldbcommon-daily',
        diamond: 'commonconfig.taobao.net:8080',
        namespace: 171,
      });
      tair._handleClusterConfig('', '', '');
      tair._handleClusterConfig('', '', '(12');
      tair._handleClusterConfig('', '', JSON.stringify({
        clusters: []
      }));
      tair._handleClusterConfig('', '', JSON.stringify({
        clusters: [{}]
      }));
      tair._handleClusterConfig('', '', JSON.stringify({
        clusters: [{address: { master: ''}}]
      }));
      tair._handleClusterConfig('', '', fs.readFileSync(path.join(fixtures, '2_clusters.txt'), 'utf8'));
    });
  });

});
