/*!
 * node-tair - test/connection.test.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var should = require('should');
var pedding = require('pedding');
var Connection = require('../lib/connection');
var Ping = require('../lib/packets/ping');
var cons = require('../lib/const');
var Transcoder = require('../lib/packets/transcoder');
var Get = require('../lib/packets/get');
var Remove = require('../lib/packets/remove');

function createConnection() {
  var connection = new Connection({
    // 10.235.145.79:5192
    host: '10.235.145.79',
    port: 5192
  });
  return connection;
}

describe('connection.test.js', function () {

  var connection;

  before(function (done) {
    connection = createConnection().on('connect', done);
  });

  after(function () {
    connection.close();
  });

  describe('call()', function () {
    it('should send a java get call', function (done) {
      var transcoder = new Transcoder();
      var get = new Get('node-tair-foo-java', transcoder, {
        namespace: 171
      });
      connection.call(get, function (err, res) {
        should.not.exist(err);
        res.should.have.keys('reqId', 'reqSize', 'packetCode', 'bodySize', 'body', 'use');
        res.body.should.have.keys('version', 'code', 'count', 'key', 'value', 'meta');

        res.body.key.should.equal('node-tair-foo-java');
        res.body.value.should.equal('java-value');
        done();
      });
    });

    it('should send a ping success', function (done) {
      connection.call(new Ping(0, 100), function (err, res) {
        should.not.exist(err);
        res.should.have.keys('reqId', 'reqSize', 'packetCode', 'bodySize', 'body', 'use');
        res.packetCode.should.equal(cons.TAIR_RESP_RETURN_PACKET);
        res.bodySize.should.equal(12);
        res.body.should.have.keys('version', 'message', 'code');
        // res.body.version.should.equal(125);
        res.body.code.should.equal(100);
        res.body.message.should.equal('');
        res.use.should.above(1);
        done();
      });
    });

    // TODO: why?
    it('should send a ping with 133 value return timeout', function (done) {
      var ping = new Ping(0, 133);
      connection.call(ping, 100, function (err, res) {
        should.exist(err);
        err.name.should.equal('TairPingRequestTimeoutError');
        err.message.should.include('Tair server no response in 100 ms, conn#');
        res.should.have.keys('reqId', 'reqSize', 'packetCode', 'bodySize', 'body', 'use');
        res.packetCode.should.equal(0);
        res.bodySize.should.equal(0);
        res.use.should.above(99);
        done();
      });
    });

    it('should send batch pings request and got the responses', function (_done) {
      var count = 1000;
      var doneCounter = 1000;
      var wrongs = [];
      var done = pedding(count, function () {
        console.log('wrong order items count: %d, first: %j', wrongs.length, wrongs[0]);
        _done();
      });

      var run = function (i) {
        var ping = new Ping(i, i);
        connection.call(ping, function (err, res) {
          doneCounter++;
          should.not.exist(err);
          res.should.have.keys('reqId', 'reqSize', 'packetCode', 'bodySize', 'body', 'use');
          res.packetCode.should.equal(cons.TAIR_RESP_RETURN_PACKET);
          res.bodySize.should.equal(12);
          res.body.should.have.keys('version', 'message', 'code');
          res.body.version.should.a.number;
          res.body.code.should.equal(i);
          // res.body.value.should.equal(doneCounter); // 此处检测不通过, 代表返回结果不一定是顺序的
          if (res.body.code !== doneCounter) {
            wrongs.push([doneCounter, i, res]);
          }
          res.use.should.above(1);
          done();
        });
      };
      for (var i = 1000; i < count + 1000; i++) {
        var index = i + 1;
        run(index);
      }
    });
  });

  describe('Request.encode(), decode()', function () {
    it('should decode wrong packet', function (done) {
      var transcoder = new Transcoder();
      var remove = new Remove('node-tair-invalid-test', transcoder, {namespace: 171});
      var req = new Connection.Request('mockconnect', remove, 10, {debug: true});
      req.id = 3;
      req.on('finish', function (err, res) {
        should.exist(err);
        if (process.version.indexOf('v0.8.') === 0) {
          err.name.should.equal('TairRemoveDecodeAssertionError');
        } else {
          err.name.should.equal('TairRemoveDecodeRangeError');
        }

        // err.message.should.equal('index out of range');
        err.data.should.have.keys('connectionName', 'reqId', 'resBuffer');
        err.data.resBuffer.should.equal('<ByteBuffer 00 01 02>');
        done();
      });
      req.decode(new Buffer([0, 1, 2]));
    });

    it('should encode remove packet', function (done) {
      var chid = 3;
      var javaBytes = new Buffer([
        109, 66, 100, 84, 0, 0, 0, 3, 0, 0, 0, 3, 0, 0, 0, 71, 0, 0,
        -85, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24, 0, 4, 110, 111, 100, 101, 45, 116,
        97, 105, 114, 45, 105, 110, 118, 97, 108, 105, 100, 45, 116, 101, 115, 116
      ]);

      var transcoder = new Transcoder();
      var remove = new Remove('node-tair-invalid-test', transcoder, {namespace: 171});
      var req = new Connection.Request('mockconnect', remove);
      req.id = chid;
      var out = {
        write: function (buf) {
          buf.should.eql(javaBytes);
          done();
        }
      };
      req.encode(out);
    });
  });

});
