/*!
 * node-tair - test/packets/prefix_puts.test.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * MIT Licensed
 *
 * Authors:
 *   fengmk2 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var fs = require('fs');
var should = require('should');
var path = require('path');
var ByteBuffer = require('byte');
var PrefixPuts = require('../../lib/packets/prefix_puts');
var Transcoder = require('../../lib/packets/transcoder');
var MixedKey = require('../../lib/packets/mixed_key');
var cons = require('../../lib/const');

var fixtures = path.join(path.dirname(__dirname), 'fixtures');

describe('packets/prefix_puts.test.js', function () {
  it('should decode "-3986: write not on master" error response', function () {
    var transcoder = new Transcoder();
    var puts = new PrefixPuts('123', [['k1', 'v1']], transcoder, {namespace: 171});
    var res = puts.decode(ByteBuffer.wrap(fs.readFileSync(path.join(fixtures, 'prefix_puts_-3986.bin'))));
    res.should.eql({
      version: 28,
      code: -3986,
      message: '',
      count: 1,
      codes: { '435674456\u0000\u0004user_interest_group_ex': -3986 }
    });
  });

  it('should decode "-3997: version error" error response', function () {
    var transcoder = new Transcoder();
    var puts = new PrefixPuts('123', [['k1', 'v1']], transcoder, {namespace: 171});
    var res = puts.decode(ByteBuffer.wrap(fs.readFileSync(path.join(fixtures, 'prefix_puts_-3997.bin'))));
    res.should.eql({
      version: 28,
      code: -3997,
      message: '',
      count: 1,
      codes: { '435674456\u0000\u0004user_interest_group_ex': -3997 }
    });
  });

  it('should encode prefix_puts bytes match java bytes', function () {
    var bodyLength = 414;
    var pkey = 'nodejs-prefixPuts-test';
    var keyValues = [
      ['k2', 2],
      ['k1', '1'],
      ['k3', 'k3']
    ];
    var chid = 2;
    var namespace = 171;
    var javaBytes = new Buffer([109, 66, 100, 84, 0, 0, 0, 2, 0, 0, 0, 24, 0, 0, 1, -98, 0, 0, -85, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24, 0, 4, 110, 111, 100, 101, 106, 115, 45, 112, 114, 101, 102, 105, 120, 80, 117, 116, 115, 45, 116, 101, 115, 116, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, -128, 0, 30, 0, 24, 0, 4, 110, 111, 100, 101, 106, 115, 45, 112, 114, 101, 102, 105, 120, 80, 117, 116, 115, 45, 116, 101, 115, 116, 0, 4, 107, 50, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 2, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, -128, 0, 30, 0, 24, 0, 4, 110, 111, 100, 101, 106, 115, 45, 112, 114, 101, 102, 105, 120, 80, 117, 116, 115, 45, 116, 101, 115, 116, 0, 4, 107, 49, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 4, 49, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, -128, 0, 30, 0, 24, 0, 4, 110, 111, 100, 101, 106, 115, 45, 112, 114, 101, 102, 105, 120, 80, 117, 116, 115, 45, 116, 101, 115, 116, 0, 4, 107, 51, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 4, 107, 51]);

    var transcoder = new Transcoder();
    var puts = new PrefixPuts(pkey, keyValues, transcoder, {namespace: 171});

    var b = puts.encode(chid);
    b.length.should.equal(javaBytes.length);
    b.should.eql(javaBytes);
    // b.slice(0, 20).should.eql(javaBytes.slice(0, 20));
  });
});
