/**!
 * node-tair - test/servers/config_server.test.js
 *
 * Copyright(c) 2013 Alibaba Group Holding Limited.
 * Authors:
 *   苏千 <suqian.yf@taobao.com> (http://fengmk2.github.com)
 */

"use strict";

/**
 * Module dependencies.
 */

var should = require('should');
var Tair = require('../../');
var ConfigServer = require('../../lib/servers/config_server');
var DataNode = require('../../lib/servers/data_node');
var accesslog = require('../../').middlewares.accesslog;

var bigValue = '大文本数据 big data大文本数据 big data大文本数据 big data大文本数据 big data大文本数据 big data大文本数据 big data大文本数据 big data大文本数据 big data大文本数据 big data大文本数据 big data大文本数据 big data';

describe('servers/config_server.test.js', function () {
  var tair;
  var aliveTair;
  var dataNode;
  var configServer;
  before(function (done) {
    var cs = configServer = new ConfigServer();
    cs.listen(6198, function () {
      tair = Tair({
        group: 'testgroup',
        servers: [{host: '127.0.0.1', port: 6198}],
      });
      tair.on('connect', done);
    });
  });

  it('should get a value return error when no alive data nodes', function (done) {
    tair.get('foo', function (err, value, res) {
      should.exist(err);
      err.name.should.equal('TairDataNodeNotAliveError');
      done();
    });
  });

  it('should get a value success when data node alive', function (done) {
    var dn = dataNode = new DataNode({
      configServers: [{host: '127.0.0.1', port: 6198}],
      group: 'testgroup',
      withHeader: false,
      host: '127.0.0.1'
    });
    dn.use(accesslog({logger: {
      accessLog: function () {
        console.log(Array.prototype.join.call(arguments, ' '));
      }
    }}));
    dn.get(500, function (req, res, next) {
      if (req.data.key === '500') {
        return next({message: 'mock error', code: -3999, status: 500});
      }
      if (req.data.key === '503') {
        return next({message: 'error to status 200', key: req.data.key});
      }
      if (req.data.key === '404') {
        return next();
      }
      if (req.data.key === 'long') {
        return res.send({value: bigValue, key: 'long'});
      }
      setTimeout(function () {
        var size = res.send({value: 'value from ' + req.data.namespace, key: req.data.key}, {use: 4, cache: 1});
        size.should.equal(125);
      }, 5);
    });
    dn.use(function (err, req, res, next) {
      if (err && err.key === '503') {
        return res.send('ignore 503');
      }
      next(err);
    });

    dn.listen(6192, function (err) {
      should.not.exist(err);
      setTimeout(function () {
        // reconnect again
        aliveTair = Tair({
          group: 'testgroup',
          servers: [{host: '127.0.0.1', port: 6198}],
          withHeader: false,
        });
        aliveTair.on('connect', function () {
          aliveTair.get('foo', {namespace: 500}, function (err, value, res) {
            should.not.exist(err);
            value.should.equal('value from 500');
            res.body.key.should.equal('foo');
            done();
          });
        });
      }, 100);
    });
  });

  it('should unsupport error when put not support', function (done) {
    aliveTair.put('foo', 'bar', {namespace: 200}, function (err, success, res) {
      should.exist(err);
      err.name.should.equal('TairRequestError');
      err.code.should.equal(-20008);
      err.message.should.equal('-20008: unsupport packet, code 1');
      done();
    });
  });

  it('should mock server error', function (done) {
    aliveTair.get('500', {namespace: 500}, function (err, value, res) {
      should.exist(err);
      err.name.should.equal('TairRequestError');
      err.message.should.equal('-3999: server error');
      err.code.should.equal(-3999);
      done();
    });
  });

  it('should get long big data', function (done) {
    aliveTair.get('long', {namespace: 500}, function (err, value, res) {
      should.not.exist(err);
      value.should.equal(bigValue);
      done();
    });
  });

  it('should get namespace: 100 return 404', function (done) {
    aliveTair.get('100\u0000', {namespace: 100}, function (err, value, res) {
      should.exist(err);
      err.name.should.equal('TairRequestError');
      err.message.should.equal('-20008: not support');
      err.code.should.equal(-20008);
      done();
    });
  });

  it('should ignore 503 error', function (done) {
    aliveTair.get('503', {namespace: 500}, function (err, value, res) {
      should.not.exist(err);
      value.should.equal('ignore 503');
      done();
    });
  });

  it('should mock 404', function (done) {
    aliveTair.get('404\u0000', {namespace: 500}, function (err, value, res) {
      should.exist(err);
      err.name.should.equal('TairRequestError');
      err.message.should.equal('-20008: not support');
      err.code.should.equal(-20008);
      done();
    });
  });
});
